""" An acoustic model with a LSTM/CTC architecture. """

import os
import numpy as np
import tensorflow as tf

from . import model

def lstm_cell(hidden_size):
    """ Wrapper function to create an LSTM cell. """

    return tf.contrib.rnn.LSTMCell(
        hidden_size, use_peepholes=True, state_is_tuple=True)

class Model(model.Model):
    """ An acoustic model with a LSTM/CTC architecture. """

    def write_desc(self):
        """ Writes a description of the model to the exp_dir. """

        path = os.path.join(self.exp_dir, "model_description.txt")
        with open(path, "w") as desc_f:
            for key, val in self.__dict__.items():
                print("%s=%s" % (key, val), file=desc_f)

    def __init__(self, exp_dir, corpus_reader, num_layers=3,
                 hidden_size=250, beam_width=100, decoding_merge_repeated=True):
        super().__init__(exp_dir, corpus_reader)

        if not os.path.isdir(exp_dir):
            os.makedirs(exp_dir)

        # Increase vocab size by 2 since we need an extra for CTC blank labels
        # and another extra for dynamic padding with zeros.
        vocab_size = corpus_reader.corpus.vocab_size+2

        # Reset the graph.
        tf.reset_default_graph()

        self.num_layers = num_layers
        self.hidden_size = hidden_size
        self.beam_width = beam_width
        self.vocab_size = vocab_size

        # Initialize placeholders for feeding data to model.
        self.batch_x = tf.placeholder(
                tf.float32, [None, None, corpus_reader.corpus.num_feats])
        self.batch_x_lens = tf.placeholder(tf.int32, [None])
        self.batch_y = tf.sparse_placeholder(tf.int32)

        batch_size = tf.shape(self.batch_x)[0]

        layer_input = self.batch_x

        for i in range(num_layers):

            with tf.variable_scope("layer_%d" % i):

                cell_fw = lstm_cell(self.hidden_size)
                cell_bw = lstm_cell(self.hidden_size)

                (self.out_fw, self.out_bw), _ = tf.nn.bidirectional_dynamic_rnn(
                        cell_fw, cell_bw, layer_input, self.batch_x_lens, dtype=tf.float32,
                        time_major=False)

                # Self outputs now becomes [batch_num, time, hidden_size*2]
                self.outputs_concat = tf.concat((self.out_fw, self.out_bw), 2)

                # For feeding into the next layer
                layer_input = self.outputs_concat

        self.outputs = tf.reshape(self.outputs_concat, [-1, self.hidden_size*2])

        # Single-variable names are appropriate for weights an biases.
        # pylint: disable=invalid-name
        W = tf.Variable(tf.truncated_normal([hidden_size*2, vocab_size],
                stddev=np.sqrt(2.0 / (2*hidden_size))))
        b = tf.Variable(tf.zeros([vocab_size]))
        self.logits = tf.matmul(self.outputs, W) + b
        self.logits = tf.reshape(self.logits, [batch_size, -1, vocab_size])
        # igormq made it time major, because of an optimization in ctc_loss.
        self.logits = tf.transpose(self.logits, (1, 0, 2))

        # For lattice construction
        self.log_softmax = tf.nn.log_softmax(self.logits)

        self.decoded, self.log_prob = tf.nn.ctc_beam_search_decoder(
                self.logits, self.batch_x_lens, beam_width=beam_width,
                merge_repeated=decoding_merge_repeated)

        # If we want to do manual PER decoding. The decoded[0] beans the best
        # hypothesis (0th) in an n-best list.
        self.dense_decoded = tf.sparse_tensor_to_dense(self.decoded[0])
        self.dense_ref = tf.sparse_tensor_to_dense(self.batch_y)

        self.loss = tf.nn.ctc_loss(self.batch_y, self.logits, self.batch_x_lens,
                preprocess_collapse_repeated=False, ctc_merge_repeated=True)
        self.cost = tf.reduce_mean(self.loss)
        self.optimizer = tf.train.AdamOptimizer().minimize(self.cost)

        self.ler = tf.reduce_mean(tf.edit_distance(
                tf.cast(self.decoded[0], tf.int32), self.batch_y))

        self.write_desc()



""" Some functions to interface with the Pangloss """

from xml.etree import ElementTree

def get_sents_times_and_translations(xml_fn):
    """ Given an XML filename, loads the transcriptions, their start/end times,
    and translations. """

    tree = ElementTree.parse(xml_fn)
    root = tree.getroot()
    if "WORDLIST" in root.tag or root.tag == "TEXT":
        transcriptions = []
        times = []
        translations = []
        for child in root:
            # If it's a sentence or a word (this will be dictated by root_tag)
            if child.tag == "S" or child.tag == "W":
                forms = child.findall("FORM")
                if len(forms) > 1:
                    # Assuming there is kindOf="phono" available.
                    for form in forms:
                        if form.attrib["kindOf"] == "phono":
                            transcription = form.text
                else:
                    transcription = child.find("FORM").text
                audio_info = child.find("AUDIO")
                if audio_info != None:
                    start_time = float(audio_info.attrib["start"])
                    end_time = float(audio_info.attrib["end"])
                    time = (start_time, end_time)
                    translation = [trans.text for trans in child.findall("TRANSL")]
                    transcriptions.append(transcription)
                    times.append(time)
                    translations.append(translation)

        return root.tag, transcriptions, times, translations
    print(root.tag)
    assert False

def remove_content_in_brackets(sentence, brackets="[]"):
    out_sentence = ''
    skip_c = 0
    for c in sentence:
        if c == brackets[0]:
            skip_c += 1
        elif c == brackets[1] and skip_c > 0:
            skip_c -= 1
        elif skip_c == 0:
            out_sentence += c
    return out_sentence



""" Interface to the Chatino data."""

from os.path import join
import os
import random
from shutil import copyfile

from .. import config
from .. import corpus
from .. import feat_extract
from .. import utils

### Initialize path variables.###
# Path to unpreprocessed Chatino data from GORILLA.
ORG_DIR = config.CHATINO_DIR
# The directory for storing input features and targets.
TGT_DIR = os.path.join(config.TGT_DIR, "chatino")
ORG_WAV_DIR = os.path.join(ORG_DIR, "wav")
# TODO Consider factoring out as non-Chatino specific
FEAT_DIR = os.path.join(TGT_DIR, "feat")
ORG_TRANSCRIPT_DIR = os.path.join(ORG_DIR, "transcriptions")
# TODO Consider factoring out as non-Chatino specific
LABEL_DIR = os.path.join(TGT_DIR, "label")

# Obtain the filename prefixes that identify recordings and their
# transcriptions
# TODO Consider factoring out as non-Chatino specific
def get_prefixes():
    return [os.path.splitext(fn)[0]
            for fn in os.listdir(ORG_TRANSCRIPT_DIR)
            if fn.endswith(".txt")]


# Hardcode the phoneme set
ONE_CHAR_PHONEMES = set(["p", "t", "d", "k", "q", "s", "x", "j", "m", "n", "r", "l",
                         "y", "w", "i", "u", "e", "o", "a"])
TWO_CHAR_PHONEMES = set(["ty", "dy", "kw", "ts", "dz", "ch", "ny", "ly",
                         "in", "en", "On", "an"])
PHONEMES = ONE_CHAR_PHONEMES.union(TWO_CHAR_PHONEMES)

# TODO Consider factoring out as non-Chatino specific
PHONEMES_TO_INDICES = {phn: index for index, phn in enumerate(PHONEMES)}
INDICES_TO_PHONEMES = {index: phn for index, phn in enumerate(PHONEMES)}

# Hardcode the tone set
ONE_CHAR_TONES = set(["0", "3", "1", "4", "2"])
TWO_CHAR_TONES = set(["04", "14", "24", "20", "42", "32", "40", "10"])
THREE_CHAR_TONES = set(["140"])
TONES = ONE_CHAR_TONES.union(TWO_CHAR_TONES.union(THREE_CHAR_TONES))

def process_transcript(org_transcript_fn, label_fn, label_type):
    """ Splits the Chatino transcript into phonemes. """

    def remove_punctuation(line):
        """ Replace symbols we don't care about.""" 

        line = line.lower()
        line = line.replace(",", "")
        line = line.replace("-", "")
        line = line.replace(".", "")
        line = line.replace("!", "")
        line = line.replace("˜", "")
        line = line.replace("+", "")
        line = line.replace(":", "")
        line = line.replace("?", "")
        line = line.replace("[blurry]", "")
        line = line.replace("[plata]", "")

        return line

    def bad_transcript(line):
        """ True if we should disregard the utterance; false otherwise."""

        if "g" in line:
            return True
        if "ñ" in line:
            return True
        if "<" in line:
            return True
        if ">" in line:
            return True
        if "b" in line:
            return True
        if "(" in line:
            return True
        if "[" in line:
            return True
        if "õ" in line:
            return True
        if "chaqF" in line:
            return True
        if "Bautista" in line:
            return True
        if "ntsaqG" in line:
            return True
        return False

    def extract_labels(line):

        # Do we want phonemes, tones, or both?
        if label_type == "phonemes":
            phonemes = True
            tones = False
        elif label_type == "tones":
            phonemes = False
            tones = True
        elif label_type == "phonemes_and_tones":
            phonemes = True
            tones = True

        words = line.split()
        labels = []
        for word in words:
            i = 0
            while i < len(word):
                if word[i:i+3] in THREE_CHAR_TONES:
                    if tones:
                        labels.append(word[i:i+3])
                    i += 3
                    continue
                if word[i:i+2] in TWO_CHAR_PHONEMES:
                    if phonemes:
                        labels.append(word[i:i+2])
                    i += 2
                    continue
                if word[i:i+2] in TWO_CHAR_TONES:
                    if tones:
                        labels.append(word[i:i+2])
                    i += 2
                    continue
                elif word[i:i+1] in ONE_CHAR_PHONEMES:
                    if phonemes:
                        labels.append(word[i:i+1])
                    i += 1
                    continue
                elif word[i:i+1] in ONE_CHAR_TONES:
                    if tones:
                        labels.append(word[i:i+1])
                    i += 1
                    continue
                elif word[i:i+1] == "'":
                    # Then assume it's distinguishing n'y from ny.
                    i += 1
                    continue
                elif word[i:i+1] == "c":
                    #It wasn't ch, return nothing and remove utterance.
                    return []
                elif word[i:i+1] == "h":
                    #It wasn't ch, return nothing and remove utterance.
                    return []
                else:
                    print(org_transcript_fn)
                    print(words)
                    print(word)
                    print(word[i:], end="")
                    input()
                    break
        return labels

    with open(org_transcript_fn) as org_f:
        line = org_f.readline()
    if bad_transcript(line):
        return
    line = remove_punctuation(line)
    phonemes = extract_labels(line)
    if phonemes == []: # If the transcription is empty, don't use the utterance
        return
    with open(label_fn, "w") as tgt_f:
        print(" ".join(phonemes), file=tgt_f)

def prepare_labels(label_type):
    """ Prepare the neural network output targets."""

    if not os.path.isdir(LABEL_DIR):
        os.makedirs(LABEL_DIR)

    for prefix in get_prefixes():
        org_fn = os.path.join(ORG_TRANSCRIPT_DIR, "%s.txt" % prefix)
        label_fn = os.path.join(
            LABEL_DIR, "%s.%s" % (prefix, label_type))
        process_transcript(org_fn, label_fn, label_type)

# TODO Consider factoring out as non-Chatino specific
def prepare_feats(feat_type):
    """ Prepare the input features."""

    if not os.path.isdir(FEAT_DIR):
        os.makedirs(FEAT_DIR)

    if feat_type=="phonemes_onehot":
        import numpy as np
        prepare_labels("phonemes")
        for prefix in get_prefixes():
            label_fn = os.path.join(LABEL_DIR, "%s.phonemes" % prefix)
            try:
                with open(label_fn) as label_f:
                    labels = label_f.readlines()[0].split()
            except FileNotFoundError:
                continue
            indices = [PHONEMES_TO_INDICES[label] for label in labels]
            one_hots = one_hots = [[0]*len(PHONEMES) for _ in labels]
            for i, index in enumerate(indices):
                one_hots[i][index] = 1
                one_hots = np.array(one_hots)
                np.save(os.path.join(FEAT_DIR, "%s.phonemes_onehot" %  prefix),
                        one_hots)
    else:
        # Otherwise, 
        for prefix in get_prefixes():
            # Convert the wave to 16k mono.
            org_wav_fn = os.path.join(ORG_WAV_DIR, "%s.wav" % prefix)
            mono16k_wav_fn = os.path.join(FEAT_DIR, "%s.wav" % prefix)
            if not os.path.isfile(mono16k_wav_fn):
                feat_extract.convert_wav(org_wav_fn, mono16k_wav_fn)

        # Extract features from the wavs.
        feat_extract.from_dir(FEAT_DIR, feat_type=feat_type)

class Corpus(corpus.AbstractCorpus):
    """ Class to interface with the Chatino corpus."""

    # TODO Reconsider the place of these splits. Perhaps train/dev/test
    # directories should be used instead, and generated in the prepare() step.
    TRAIN_VALID_TEST_SPLIT = [2048, 207, 206]
    FEAT_DIR = FEAT_DIR
    LABEL_DIR = LABEL_DIR

    def __init__(self, feat_type, label_type, max_samples=1000):
        super().__init__(feat_type, label_type)

        if label_type == "phonemes":
            self.labels = PHONEMES
        elif label_type == "tones":
            self.labels = TONES
        elif label_type == "phonemes_and_tones":
            self.labels = PHONEMES.union(TONES)
        else:
            raise ValueError("label_type=%s not supported." % (label_type))

        self.feat_type = feat_type
        self.label_type = label_type

        # Filter prefixes based on what we find in the feat/ dir.
        self.prefixes = [prefix for prefix in get_prefixes()
                         if os.path.isfile(os.path.join(
                             FEAT_DIR, "%s.%s.npy" % (prefix, feat_type)))]
        # Filter prefixes based on what we find in the label/ dir.
        self.prefixes = [prefix for prefix in self.prefixes
                         if os.path.isfile(os.path.join(
                             LABEL_DIR, "%s.%s" % (prefix, label_type)))]

        # Remove prefixes whose feature files are too long.
        if max_samples:
            self.prefixes = utils.filter_by_size(
                FEAT_DIR, self.prefixes, feat_type, max_samples)

        random.seed(0)
        random.shuffle(self.prefixes)
        self.train_prefixes = self.prefixes[:self.TRAIN_VALID_TEST_SPLIT[0]]
        valid_end = self.TRAIN_VALID_TEST_SPLIT[0]+self.TRAIN_VALID_TEST_SPLIT[1]
        self.valid_prefixes = self.prefixes[self.TRAIN_VALID_TEST_SPLIT[0]:valid_end]
        self.test_prefixes = self.prefixes[valid_end:]

        # Writing validation set prefixes so that latex_output can give
        # meaningful output.
        with open(join(TGT_DIR, "valid_prefixes.txt"), "w") as valid_f:
            for prefix in self.valid_prefixes:
                print(prefix, file=valid_f)

        # TODO Make the distinction between this and the constants at the start
        # of the file clear.
        self.LABEL_TO_INDEX = {label: index for index, label in enumerate(
                                 ["pad"] + sorted(list(self.labels)))}
        self.INDEX_TO_LABEL = {index: label for index, label in enumerate(
                                 ["pad"] + sorted(list(self.labels)))}
        self.vocab_size = len(self.labels)

    # TODO Use 'labels' instead of 'phonemes' here and in corpus.py
    # Also, factor out as non-Chatino-specific.
    def indices_to_phonemes(self, indices):
        return [(self.INDEX_TO_LABEL[index]) for index in indices]
    def phonemes_to_indices(self, labels):
        return [self.LABEL_TO_INDEX[label] for label in labels]



""" Script for formatting results of experiments """

from . import config
from .datasets import na
from .datasets import chatino

from collections import defaultdict
import os
from . import utils

from distance import min_edit_distance_align
from distance import cluster_alignment_errors

def round_items(floats):
    return ["%0.3f" % fl for fl in floats]

def average(exp_nums, show_tgm_f1=False, phones=na.PHONEMES, tones=na.TONES):
    """ Averages the results across a few experimental runs. """

    ler_total = 0
    per_total = 0
    ter_total = 0
    tgm_f1_total = 0
    for i, exp_num in enumerate(exp_nums):
        path = os.path.join(config.EXP_DIR, str(exp_num))
        ler, per, ter = test_results(path, phones, tones)
        tgm_f1 = 0
        if show_tgm_f1:
            tgm_f1 = symbol_f1(exp_num, "|")
        print("Exp #{}:".format(i))
        print("\tPER & TER & LER & TGM-F1")
        print("\t{} & {} & {} & {}".format(per, ter, ler, tgm_f1))
        ler_total += ler
        per_total += per
        ter_total += ter
        tgm_f1_total +=  tgm_f1
    print("Average:")
    print("\tPER & TER & LER & TGM-F1")
    print("\t{} & {} & {} & {}\\\\".format(
        per_total/(i+1), ter_total/(i+1), ler_total/(i+1), tgm_f1_total/(i+1)))

def format(exp_paths,
                   phones=na.PHONEMES,
                   tones=na.TONES):
    """ Takes a list of experimental paths such as mam/exp/<number> and outputs
    the results. """

    valid_lers = []
    valid_pers = []
    test_lers = []
    test_pers = []
    test_ters = []

    for path in exp_paths:

        test_ler, test_per, test_ter = test_results(path, phones, tones)
        test_lers.append(test_ler)
        test_pers.append(test_per)
        test_ters.append(test_ter)

        with open(os.path.join(path, "best_scores.txt")) as best_f:
            sp = best_f.readline().replace(",", "").split()
            training_ler, valid_ler, valid_per = float(sp[4]), float(sp[7]), float(sp[10])
            valid_lers.append(valid_ler)

    print("Valid LER", round_items(valid_lers))
    print("Test LER", round_items(test_lers))
    print("Test PER", round_items(test_pers))
    print("Test TER", round_items(test_ters))

    print("PERS:")
    for item in zip([128,256,512,1024,2048], test_pers):
        print("(%d, %f)" % item)

    print("TERS:")
    for item in zip([128,256,512,1024,2048], test_ters):
        print("(%d, %f)" % item)

def filter_labels(sent, labels=None):
    """ Returns only the tokens present in the sentence that are in labels."""
    if labels:
        return [tok for tok in sent if tok in labels]
    return sent

def test_results(exp_path, phones, tones):
    """ Gets results of the model on the test set. """

    test_path = os.path.join(exp_path, "test")
    print(test_path)
    with open(os.path.join(test_path, "test_per")) as test_f:
        line = test_f.readlines()[0]
        test_ler = float(line.split()[2].strip(","))

    test_per = filtered_error_rate(os.path.join(test_path, "hyps"),
                                      os.path.join(test_path, "refs"),
                                      phones)

    test_ter = filtered_error_rate(os.path.join(test_path, "hyps"),
                                      os.path.join(test_path, "refs"),
                                      tones)

    return test_ler, test_per, test_ter

def filtered_error_rate(hyps_path, refs_path, labels):

    with open(hyps_path) as hyps_f:
        lines = hyps_f.readlines()
        hyps = [filter_labels(line.split(), labels) for line in lines]
    with open(refs_path) as refs_f:
        lines = refs_f.readlines()
        refs = [filter_labels(line.split(), labels) for line in lines]

    # For the case where there are no tokens left after filtering.
    only_empty = True
    for entry in hyps:
        if entry != []:
            only_empty = False
    if only_empty:
        return -1

    return utils.batch_per(hyps, refs)

def ed_alignments(exp_path):

    test_path = os.path.join(exp_path, "test")
    hyps_path = os.path.join(test_path, "hyps")
    refs_path = os.path.join(test_path, "refs")

    with open(hyps_path) as hyps_f:
        lines = hyps_f.readlines()
        hyps = [line.split() for line in lines]
    with open(refs_path) as refs_f:
        lines = refs_f.readlines()
        refs = [line.split() for line in lines]

    alignments = []
    for ref, hyp in zip(refs, hyps):
        alignment = min_edit_distance_align(ref, hyp)
        alignments.append(alignment)

    return alignments

def symbol_f1(exp_num, symbol):

    exp_path = os.path.join(config.EXP_DIR, str(exp_num))
    alignments = ed_alignments(exp_path)

    correct = 0
    del_ = 0
    ins = 0
    del_sub = 0
    ins_sub = 0
    total_hyp = 0
    total_ref = 0

    tp = 0
    fp = 0
    fn = 0
    tn = 0
    for alignment in alignments:
        for arrow in alignment:
            if arrow[0] == symbol:
                if arrow[1] == symbol:
                    tp +=1
                else:
                    fn += 1
            elif arrow[0] != symbol:
                if arrow[1] == symbol:
                    fp += 1
                else:
                    tn += 1

    sensitivity = tp/(tp+fn)
    specificity = tn/(tn+fp)
    accuracy = (tp + tn)/(tp + tn + fp + fn)
    precision = tp/(tp+fp)
    fdr = fp/(tp+fp)
    f1 = 2*tp/(2*tp + fp + fn) # precision and sensitivity harmonic mean.
    #print("Sensitivity: {}".format(sensitivity))
    #print("(False negative rate: {})".format(1-sensitivity))
    #print("Specificity: {}".format(specificity))
    #print("(False positive rate: {})".format(1-specificity))
    #print("Accuracy: {}".format(accuracy))
    #print("(Error rate: {})".format(1-accuracy))
    #print("Precision: {}".format(precision))
    #print("(False discovery rate: {})".format(fdr))
    #print("F1: {}".format(f1))
    #print("TP: {}, FN: {}".format(tp, fn))
    #print("TN: {}, FP: {}".format(tn, fp))

    errors = []
    for alignment in alignments:
        for arrow in alignment:
            if arrow[0] != arrow[1]:
                errors.append(arrow)

    err_hist = {}
    for error in errors:
        if error in err_hist:
            err_hist[error] += 1
        else:
            err_hist[error] = 1

    return f1

def latex_output(refs_paths, hyps_paths, utter_ids_fn):
    """ Pretty print the hypotheses and references. """

    alignment_matrix = []
    for refs_path, hyps_path in zip(refs_paths, hyps_paths):
        with open(refs_path) as refs_f:
            lines = refs_f.readlines()
            refs = [line.split() for line in lines]

        with open(hyps_path) as hyps_f:
            lines = hyps_f.readlines()
            hyps = [line.split() for line in lines]

        alignments = []
        for ref, hyp in zip(refs, hyps):
            alignment = min_edit_distance_align(ref, hyp)
            alignments.append(alignment)
        alignment_matrix.append(alignments)

    with open(utter_ids_fn) as f:
        prefixes = [line.replace("_", "\_") for line in f]
        # TODO Na-specific stuff commented out for work on Chatino
        #prefixes2 = []
        #for prefix in prefixes:
        #    sp = prefix.split(".")
        #    prefixes2.append(" ".join([sp[0], "Sent \\#" + str(int(sp[1])+1)]))
        #prefixes = prefixes2
    #print(prefixes)

    with open("hyps_refs.tex", "w") as out_f:

        print("\documentclass[10pt]{article}\n"
              "\\usepackage[a4paper,margin=0.5in,landscape]{geometry}\n"
              "\\usepackage[utf8]{inputenc}\n"
              "\\usepackage{xcolor}\n"
              "\\usepackage{polyglossia}\n"
              "\\usepackage{booktabs}\n"
              "\\usepackage{longtable}\n"
              "\setmainfont[Mapping=tex-text,Ligatures=Common,Scale=MatchLowercase]{Doulos SIL}\n"
              "\DeclareRobustCommand{\hl}[1]{{\\textcolor{red}{#1}}}\n"
              "% Hyps path: " + hyps_path +
              "\\begin{document}\n"
              "\\begin{longtable}{ll}", file=out_f)

        print("\\toprule", file=out_f)
        for sent in zip(prefixes, *alignment_matrix):
            prefix = sent[0]
            alignments = sent[1:]
            print("Utterance ID: &", prefix.strip(), "\\\\", file=out_f)
            for i, alignment in enumerate(alignments):
                ref_list = []
                hyp_list = []
                for arrow in alignment:
                    if arrow[0] == arrow[1]:
                        # Then don't highlight it; it's correct.
                        ref_list.append(arrow[0])
                        hyp_list.append(arrow[1])
                    else:
                        # Then highlight the errors.
                        ref_list.append("\hl{%s}" % arrow[0])
                        hyp_list.append("\hl{%s}" % arrow[1])
                print("Ref: &", "".join(ref_list), "\\\\", file=out_f)
                print("{}: &".format(hyps_paths[i].replace("_", "\_")), "".join(hyp_list), "\\\\", file=out_f)
            print("\\midrule", file=out_f)

        print("\end{longtable}", file=out_f)
        print("\end{document}", file=out_f)

def error_types(exp_path, labels=None):
    """ Stats about the most common types of errors in the test set."""

    test_path = os.path.join(exp_path, "test")
    hyps_path = os.path.join(test_path, "hyps")
    refs_path = os.path.join(test_path, "refs")

    with open(hyps_path) as hyps_f:
        lines = hyps_f.readlines()
        hyps = [filter_labels(line.split(), labels) for line in lines]
    with open(refs_path) as refs_f:
        lines = refs_f.readlines()
        refs = [filter_labels(line.split(), labels) for line in lines]

    alignments = []
    errors = []
    for ref, hyp in zip(refs, hyps):
        alignment = min_edit_distance_align(ref, hyp)
        alignment = cluster_alignment_errors(alignment)
        alignments.append(alignment)
        for arrow in alignment:
            if arrow[0] != arrow[1]:
                errors.append(arrow)

    err_hist = {}
    for error in errors:
        if error in err_hist:
            err_hist[error] += 1
        else:
            err_hist[error] = 1

    error_list = sorted(err_hist.items(), key=lambda x: x[1], reverse=False)
    for thing in error_list:
        print(thing)

    subs = 0
    inss = 0
    dels = 0
    for error in error_list:
        if len(error[0][1]) == 0:
            dels += 1
        if len(error[0][0]) == 0:
            inss += 1
        else:
            subs += 1
    print(subs)
    print(inss)
    print(dels)

def chatino_tone_confusion(exp_path):
    """ Outputs a confusion matrix for Chatino tones."""

    tones = list(chatino.TONES)
    alignments = ed_alignments(exp_path)

    d = defaultdict(int)
    for alignment in alignments:
        for arrow in alignment:
            if arrow[0] in tones:
                d[arrow] += 1

    import pprint; pprint.pprint(d)

    total = defaultdict(int)
    for ref in tones:
        for hyp in tones:
            total[ref] += d[(ref, hyp)]

    nonzero_tones = [item[0] for item in total.items() if item[1] > 4]
    print(sorted(total.items(), key=lambda x: x[1]))

    for hyp in nonzero_tones[:-1]:
        print(hyp + ",", end="")
    print("%s\\\\" % nonzero_tones[-1])
    for ref in nonzero_tones:
        print(ref + ",", end="")
        for hyp in nonzero_tones[:-1]:
            print("%0.3f," % (d[(ref, hyp)]*100/total[ref]), end="")
        print("%0.3f\\\\" % (d[(ref, nonzero_tones[-1])]*100/total[ref]))




""" Ultimately a module to produce lattices from the multinomials of CTC-based
    neural networks. Includes other methods to do things such as sample and
    collapse paths.
"""

import os
import subprocess

import tensorflow as tf

from .corpus_reader import CorpusReader

from . import config

def test_sampling():
    """ As a proof of concept, first we want to sample paths from a trained CTC
    network's multinomials. After collapsing them, we want to assess their
    phoneme error rate to ensure we frequently get decent PERs."""
    pass

def compile_fst(prefix, syms_fn):
    """ Compiles the given text-based FST into a binary using OpenFST."""

    # Compile the fst
    args = [os.path.join(config.OPENFST_BIN_PATH, "fstcompile"),
            "--arc_type=log",
            "--isymbols=%s" % syms_fn,
            "--osymbols=%s" % syms_fn,
            "%s.txt" % prefix, "%s.bin" % prefix]
    subprocess.run(args)

def draw_fst(prefix, syms_fn):
    """ Draws the given fst using dot."""

    args = ["fstdraw", "--isymbols=%s" % syms_fn,
            "--osymbols=%s" % syms_fn,
            "%s.bin" % prefix, "%s.dot" % prefix]
    subprocess.run(args)

#    args = ["dot", "-Tpdf", "%s.dot" % prefix]
#    with open("%s.pdf" % prefix, "w") as out_f:
#        subprocess.run(args, stdout=out_f)

def create_symbol_table(index_to_token, filename):
    """ Creates a symbol table for the given vocab."""

    with open(filename, "w") as out_f:
        print("<eps> 0", file=out_f)
        for phone_id, phone in index_to_token.items():
            print("%s %d" % (phone, phone_id+1), file=out_f)
        print("<bl> %d" % (len(index_to_token)+1), file=out_f)

def logsoftmax2confusion(logsoftmax, index_to_token, prefix, beam_size):
    """ Converts a sequence of softmax outputs into a confusion network."""

    with open(prefix + ".confusion.txt", "w") as out_f:
        for node_id, timestep_softmax in enumerate(logsoftmax):
            phone_probs = list(zip(timestep_softmax, range(len(timestep_softmax))))
            # Only take the top entries within the beam.
            beam = sorted(phone_probs, reverse=True)[:beam_size]
            for prob, phone_id in beam:
                if phone_id == len(index_to_token):
                    # Then it's just outside the range of the mapping and is
                    # thus a blank symbol
                    print("%d %d <bl> <bl> %f" % (node_id, node_id+1, -prob),
                          file=out_f)
                else:
                    print("%d %d %s %s %f" % (
                          node_id, node_id+1,
                          index_to_token[phone_id], index_to_token[phone_id],
                          -prob),
                          file=out_f)
        # If I'm writing -logsoftmax values where they are arc costs, then
        # I believe 0 means a cost of zero.
        print("%d 0" % (node_id+1), file=out_f)

def create_collapse_fst(index_to_token, fst_fn):
    with open(fst_fn, "w") as out_f:
        for i, phone in sorted(index_to_token.items()):
            print("0 %d %s %s" % (i+1, phone, phone), file=out_f)
            print("%d %d %s <eps>" % (i+1, i+1, phone), file=out_f)
            for j, phone in sorted(index_to_token.items()):
                if j != i:
                    print("%d %d %s %s" % (i+1, j+1, phone, phone), file=out_f)
            print("%d 0 <bl> <eps>" % (i+1), file=out_f)
        print("0 0 <bl> <eps>", file=out_f)
        for i, phone in sorted(index_to_token.items()):
            print("%d 0" % (i+1), file=out_f)
        print("0 0", file=out_f)

def logsoftmax2lattice(logsoftmax, index_to_token, prefix, beam_size):
    """ Takes a numpy array of shape [time_steps, vocab] along with a function
    to produce tokens from their indices and an output filename prefix. A
    lattice is written to the output filename.
    """

    # Create the confusion network based on logsoftmax
    logsoftmax2confusion(logsoftmax, index_to_token, prefix, beam_size)



""" Configuration for some variables that may be system dependent.

Supply a settings.ini file to override values. The format of the file will
have the following sections

[PATHS]
CORPORA_BASE_PATH = /my/path/to/corpora
TARGET = /path/to/preprocessed/data
EXPERIMENTS = /path/to/experiment/output
"""
import configparser
import os


config_file = configparser.ConfigParser()
config_file.read('settings.ini')

CORPORA_BASE_PATH = config_file.get("PATHS", "CORPORA_BASE_PATH",
                                    fallback="/lt/work/oadams/")

# The directory of the original source corpora, un-preprocessed.
CHATINO_DIR = os.path.join(CORPORA_BASE_PATH, "chatino", "CTP")
NA_DIR = os.path.join(CORPORA_BASE_PATH, "Na")
# Personal corpora files
# TODO: fix these stubs
KUNWINJKU_STEVEN_DIR = config_file.get("PATHS", "KUNWINJKU_STEVEN_PATH", fallback="")
EN_WORDS_PATH = config_file.get("PATHS", "EN_WORDS_PATH", fallback="")
BUTCHER_DIR = os.path.join(CORPORA_BASE_PATH, "OA_Data")

# The directory where the preprocessed data will be held.
TGT_DIR = config_file.get("PATHS", "TARGET", fallback="./data")
# The path for experiments
EXP_DIR = config_file.get("PATHS", "EXPERIMENTS", fallback="./exp")

# The path to the sox tool; currently used for splitting WAVs, but we can
# replace with pydub TODO
SOX_PATH = config_file.get("PATHS", "SOX_BASE_PATH", fallback="/home/oadams/tools/sox-14.4.2/src/sox")
# FFMPEG is used for normalizing WAVs
FFMPEG_PATH = config_file.get("PATHS", "FFMPEG_PATH", fallback="ffmpeg")
# Kaldi is used for pitch extraction
KALDI_ROOT = config_file.get("PATHS", "KALDI_ROOT_PATH", fallback="/home/oadams/tools/kaldi")
# Used for lattice output
OPENFST_BIN_PATH = config_file.get("PATHS", "OPEN_FST_BIN_PATH", fallback="/home/oadams/tools/openfst-1.6.2/src/bin")



""" Interface to Steven's Kunwinjku data. """

from collections import namedtuple
import glob
import os
from os.path import join

import pympi

from .. import config

ORG_DIR = config.KUNWINJKU_STEVEN_DIR
TGT_DIR = join(config.TGT_DIR, "kunwinjku-steven")

with open(config.EN_WORDS_PATH) as words_f:
    EN_WORDS = words_f.readlines()
    EN_WORDS = set([word.strip() for word in EN_WORDS])
    NA_WORDS_IN_EN_DICT = set(["Kore", "Nani", "karri", "imi", "o", "yaw", "i-",
                           "bi-", "aye", "imi", "ane", "kubba", "kab", "a-",
                           "ad", "a", "Mak", "Selim", "ngai", "en", "yo",
                           "wud", "Mani", "yak", "Manu", "ka-", "mong",
                           "manga", "ka-", "mane", "Kala", "name", "kayo",
                           "Kare", "laik", "Bale"])
    EN_WORDS = EN_WORDS.difference(NA_WORDS_IN_EN_DICT)

print("imi" in EN_WORDS)

def good_elan_paths():
    """
    Returns a list of ELAN files for recordings with good quality audio, as
    designated by Steven.
    """

    with open(join(ORG_DIR, "good-files.txt")) as path_list:
        good_paths = [path.strip() for path in path_list]

    elan_paths = []
    for path in good_paths:
        _, ext = os.path.splitext(path)
        if ext == ".eaf":
            elan_paths.append(join(ORG_DIR, path))
        else:
            full_path = join(ORG_DIR, path)
            if os.path.isdir(full_path):
                for elan_path in glob.glob('{}/**/*.eaf'.format(full_path),
                                           recursive=True):
                    elan_paths.append(elan_path)

    return elan_paths

def explore_elan_files(elan_paths):
    """
    A function to explore the tiers of ELAN files.
    """

    for elan_path in elan_paths:
        print(elan_path)
        eafob = pympi.Elan.Eaf(elan_path)
        tier_names = eafob.get_tier_names()
        for tier in tier_names:
            print("\t", tier)
            try:
                for annotation in eafob.get_annotation_data_for_tier(tier):
                    print("\t\t", annotation)
            except KeyError:
                continue

        input()

def elan_utterances():
    """ Collects utterances from various ELAN tiers. This is based on analysis
    of hte 'good' ELAN files, and may not generalize."""

    elan_tiers = {"rf", "rf@RN", "rf@MARK",
                  "xv", "xv@RN", "xv@MN", "xv@JN", "xv@EN", "xv@MARK", "xv@GN",
                  "nt@RN", "nt@JN",
                  "PRN_free", "PRN_Pfx", "NmCl_Gen", "ng_DROP",
                  "Other",
                 }

    Utterance = namedtuple("Utterance", ["file", "start_time", "end_time", "text"])

    utterances = []
    all_utts = []
    for elan_path in good_elan_paths():
        eafob = pympi.Elan.Eaf(elan_path)
        #import pprint; pprint.pprint(dir(eafob))
        can_find_path = False
        for md in eafob.media_descriptors:
            try:
                media_path = os.path.join(os.path.dirname(elan_path),
                                          md["RELATIVE_MEDIA_URL"])
                if os.path.exists(media_path):
                    # Only one media_path file is needed, as long as it exists.
                    can_find_path = True
                    break
                # Try just looking for the basename specified in the
                # RELATIVE_MEDIA_URL
                media_path = os.path.join(os.path.dirname(elan_path),
                                          os.path.basename(md["RELATIVE_MEDIA_URL"]))
                if os.path.exists(media_path):
                    can_find_path = True
                    break
            except KeyError:
                # Then it might be hard to find the MEDIA URL if its not
                # relative. Keep trying.
                continue
        if can_find_path:
            tier_names = eafob.get_tier_names()
            for tier in tier_names:
                if tier.startswith("rf") or tier.startswith("xv") or tier in elan_tiers:
                    for annotation in eafob.get_annotation_data_for_tier(tier):
                        utterance = Utterance(media_path, *annotation[:3])
                        if not utterance.text.strip() == "":
                            utterances.append(utterance)
        else:
            print("Warning: Can't find the media file for {}".format(elan_path))

    return utterances

def segment_gup_phonemes(utterance):
    """
    Takes as input a string in Kunwinjku and segments it into phoneme-like
    units based on the standard orthographic rules specified at
    http://bininjgunwok.org.au/
    """

    utterance = utterance.lower()
    print(utterance)

def explore_code_switching():

    utters = elan_utterances()

    en_count = 0
    for utter in utters:
        for word in utter.text.split():
            if word in EN_WORDS:
                en_count += 1
                print(utter.text)
                print("\t" + repr(word))
                #input()
                break
    print(en_count)
    print(len(utters))

"""
rf
rf@RN
xv (Esther_1.eaf, but has some English in it)
PRN_free, PRN_Pfx, NmCl_Gen, ng_DROP (Esther_2.eaf. Code switching might be an
	issue. Otherwise Esther files are promising for the finer granularity that
	is additionally included.)
xv@RN (Kabo.eaf, but its a duplicate of the xv tier. More pure in kunboy.eaf)
xv@MN, LEX (manbarndarr.eaf)
xv@JN, PRN_pfx, PRFX (Manbedgje.eaf)
Other (Manbulu.eaf)
nt@JN (Mandeb.eaf, but also has English in Mandjarduk_b.eaf)
xv@EN (mandjimdjim.eaf)
nt@RN (20161013_manmorlak.eaf)
rf@MARK, xv@MARK (Mark on rock. Has a few different tiers, some of which have
multiple text fields. A bit of code switching).
xv@GN (Marys_Yirlinkirrkirr.eaf)
A bunch more xv@s (Njanjma_Injalak_weaving.eaf)
Topic Index (Terrah_ngalwarrngurru.eaf)

Can just check across all these tiers and any that are empty or contain empty
strings I can just ignore.
"""



""" Generic model for automatic speech recognition. """

import inspect
import itertools
import os
from operator import itemgetter
import subprocess
import sys

import numpy as np
import tensorflow as tf

from . import utils
from . import lattice
from . import config
from .exceptions import PersephoneException

OPENFST_PATH = config.OPENFST_BIN_PATH

allow_growth_config = tf.ConfigProto()
allow_growth_config.gpu_options.allow_growth=True #pylint: disable=no-member

class Model:
    """ Generic model for our ASR tasks. """

    # Subclasses should instantiate these variables:
    exp_dir = None
    batch_x = None
    batch_x_lens = None
    batch_y = None
    optimizer = None
    ler = None
    dense_decoded = None
    dense_ref = None
    corpus_reader = None
    saved_model_path = None

    def __init__(self, exp_dir, corpus_reader):
        self.exp_dir = exp_dir
        self.corpus_reader = corpus_reader
        self.log_softmax = None

    def transcribe(self, restore_model_path=None):
        """ Transcribes an untranscribed dataset. Similar to eval() except
        no reference translation is assumed, thus no LER is calculated.
        """

        saver = tf.train.Saver()
        with tf.Session(config=allow_growth_config) as sess:
            if restore_model_path:
                saver.restore(sess, restore_model_path)
            else:
                if self.saved_model_path:
                    saver.restore(sess, self.saved_model_path)
                else:
                    raise PersephoneException("No model to use for transcription.")

            batch_gen = self.corpus_reader.untranscribed_batch_gen()

            hyp_batches = []
            for batch_i, batch in enumerate(batch_gen):

                batch_x, batch_x_lens, feat_fn_batch = batch

                feed_dict = {self.batch_x: batch_x,
                             self.batch_x_lens: batch_x_lens}

                [dense_decoded] = sess.run([self.dense_decoded], feed_dict=feed_dict)
                hyps = self.corpus_reader.human_readable(dense_decoded)

                # Prepare dir for transcription
                hyps_dir = os.path.join(self.exp_dir, "transcriptions")
                if not os.path.isdir(hyps_dir):
                    os.mkdir(hyps_dir)

                hyp_batches.append((hyps,feat_fn_batch))

            with open(os.path.join(hyps_dir, "hyps.txt"), "w") as hyps_f:
                for hyp_batch, fn_batch in hyp_batches:
                    for hyp, fn in zip(hyp_batch, fn_batch):
                        print(fn, file=hyps_f)
                        print(" ".join(hyp), file=hyps_f)
                        print("", file=hyps_f)

            """
            # TODO This sorting is Na-corpus centric and won't generalize. It
            # is to sort by recording name (ie. Benevolence) then by utterance
            # id within that (Benevolence.0, benevolence.1, ...)
            utters = [(hyps, feat_fn, feat_fn.split(".")[0], int(feat_fn.split(".")[1]))
                      for hyps, feat_fn in zip(hyps, feat_fn_batch)]
            print(utters)
            utters.sort(key=itemgetter(2,3))
            print(utters)
            with open(os.path.join(hyps_dir, "hyps"), "w") as hyps_f:
                for hyp, fn, _, _ in utters:
                    fn = "_".join(os.path.basename(fn).split(".")[:2])
                    print(fn + ": ", file=hyps_f)
                    print(" ".join(hyp), file=hyps_f)
            """

    def output_lattices(self, batch, restore_model_path=None):
        """ Outputs the logits from the model, given an input batch, so that
            lattices can ultimately be extracted."""

        saver = tf.train.Saver()
        with tf.Session(config=allow_growth_config) as sess:
            if restore_model_path:
                saver.restore(sess, restore_model_path)
            else:
                assert self.saved_model_path
                saver.restore(sess, self.saved_model_path)

            batch_x, batch_x_lens, batch_y = batch

            feed_dict = {self.batch_x: batch_x,
                         self.batch_x_lens: batch_x_lens,
                         self.batch_y: batch_y}

            # Get the log_softmax matrices
            log_softmax = sess.run([self.log_softmax], feed_dict=feed_dict)
            log_softmax = log_softmax[0]
            log_softmax = np.swapaxes(log_softmax, 0, 1)
            out_dir = os.path.join(self.exp_dir, "lattice")
            if not os.path.exists(out_dir):
                os.makedirs(out_dir)
            for i, example in enumerate(log_softmax):
                length = batch_x_lens[i]
                np.save(os.path.join(out_dir, "utterance_%d_log_softmax" % i),
                        example[:length])

        ### Create the lattices.###
        index_to_token = self.corpus_reader.corpus.INDEX_TO_PHONEME

        # Create symbol table.
        syms_fn = os.path.join(out_dir, "symbols.txt")
        lattice.create_symbol_table(index_to_token, syms_fn)

        # Create the FST that removes blanks and repeated tokens.
        lattice.create_collapse_fst(index_to_token,
                                    os.path.join(out_dir, "collapse_fst.txt"))
        lattice.compile_fst(os.path.join(out_dir, "collapse_fst"), syms_fn)

        for i, log_softmax_example in enumerate(log_softmax):
            # Create a confusion network
            length = batch_x_lens[i]
            lattice.logsoftmax2confusion(log_softmax_example[:length],
                                         index_to_token,
                                         os.path.join(out_dir, "utterance_%d" % i),
                                         beam_size=4)
            lattice.compile_fst(os.path.join(out_dir, "utterance_%d.confusion" % i),
                                syms_fn)

            prefix = os.path.join(out_dir, "utterance_%d" % i)

            try:
                run_args = [os.path.join(OPENFST_PATH, "fstarcsort"),
                            prefix + ".confusion.bin",
                            prefix + ".confusion.sort.bin"]
                subprocess.run(run_args)

                # Compose the confusion network with the FST that removes blanks
                # and repetitions, expanding to a larger lattice-like FST.
                run_args = [os.path.join(OPENFST_PATH, "fstcompose"),
                            prefix + ".confusion.sort.bin",
                            os.path.join(out_dir, "collapse_fst.bin"),
                            prefix + ".collapsed.bin"]
                subprocess.run(run_args)

                # Take the output projection of the FST.
                run_args = [os.path.join(OPENFST_PATH, "fstproject"),
                            "--project_output",
                            prefix + ".collapsed.bin", prefix + ".projection.bin"]
                subprocess.run(run_args)

                # Push weights
#               run_args = [os.path.join(OPENFST_PATH, "fstpush"),
#                           "--push_weights",
#                           prefix + ".projection.bin", prefix + ".pushed.bin"]
#               subprocess.run(run_args)

                # Remove epsilons
                run_args = [os.path.join(OPENFST_PATH, "fstrmepsilon"),
                            "--reverse=true",
                            prefix + ".projection.bin", prefix + ".rmepsilon.bin"]
                subprocess.run(run_args)
            except FileNotFoundError:
                print("Make sure you have OpenFST binaries installed and "
                      "available on the path")
                raise

    def eval(self, restore_model_path=None):
        """ Evaluates the model on a test set."""

        saver = tf.train.Saver()
        with tf.Session(config=allow_growth_config) as sess:
            if restore_model_path:
                saver.restore(sess, restore_model_path)
            else:
                assert self.saved_model_path
                saver.restore(sess, self.saved_model_path)

            test_x, test_x_lens, test_y = self.corpus_reader.test_batch()

            feed_dict = {self.batch_x: test_x,
                         self.batch_x_lens: test_x_lens,
                         self.batch_y: test_y}

            test_ler, dense_decoded, dense_ref = sess.run(
                [self.ler, self.dense_decoded, self.dense_ref],
                feed_dict=feed_dict)
            hyps, refs = self.corpus_reader.human_readable_hyp_ref(
                dense_decoded, dense_ref)
            # Log hypotheses
            hyps_dir = os.path.join(self.exp_dir, "test")
            if not os.path.isdir(hyps_dir):
                os.mkdir(hyps_dir)
            with open(os.path.join(hyps_dir, "hyps"), "w") as hyps_f:
                for hyp in hyps:
                    print(" ".join(hyp), file=hyps_f)
            with open(os.path.join(hyps_dir, "refs"), "w") as refs_f:
                for ref in refs:
                    print(" ".join(ref), file=refs_f)

            test_per = utils.batch_per(hyps, refs)
            with open(os.path.join(hyps_dir, "test_per"), "w") as per_f:
                print("Test PER: %f, tf LER: %f" % (test_per, test_ler), file=per_f)

    def train(self, early_stopping_steps=10, min_epochs=30, max_valid_ler=1.0,
              max_train_ler=0.3, max_epochs=100, restore_model_path=None):
        """ Train the model.

            batch_size: The number of utterances in each batch.
            total_size: The number of TIMIT training examples to use.
            num_epochs: The number of times to iterate over all the training
                        examples.
            feat_type:  Is the identifier for the type of features we're using.
                        'mfcc13_d' means MFCCs of 13 dimensions with their first
                        derivatives.
            save_n: Whether to save the model at every n epochs.
            restore_model_path: The path to restore a model from.
        """

        # Not technically the upper bound on a LER but we don't want to save if
        # it's not below this.
        best_valid_ler = 1.0
        steps_since_last_record = 0

        #Get information about training for the names of output files.
        frame = inspect.currentframe()
        # pylint: disable=deprecated-method
        # It was a mistake to deprecate this in Python 3.5
        args, _, _, values = inspect.getargvalues(frame)
        with open(os.path.join(self.exp_dir, "train_description.txt"), "w") as desc_f:
            for arg in args:
                if type(values[arg]) in [str, int, float] or isinstance(
                        values[arg], type(None)):
                    print("%s=%s" % (arg, values[arg]), file=desc_f)
                else:
                    print("%s=%s" % (arg, values[arg].__dict__), file=desc_f)
            print("num_train=%s" % (self.corpus_reader.num_train), file=desc_f)
            print("batch_size=%s" % (self.corpus_reader.batch_size), file=desc_f)

        out_file = open(os.path.join(self.exp_dir, "train_log.txt"), "w")

        # Load the validation set
        valid_x, valid_x_lens, valid_y = self.corpus_reader.valid_batch()

        saver = tf.train.Saver()

        sess = tf.Session(config=allow_growth_config)

        if restore_model_path:
            saver.restore(sess, restore_model_path)
        else:
            sess.run(tf.global_variables_initializer())

        # Prepare directory to output hypotheses to
        hyps_dir = os.path.join(self.exp_dir, "decoded")
        if not os.path.isdir(hyps_dir):
            os.mkdir(hyps_dir)

        best_epoch_str = None
        for epoch in itertools.count():
            print("\nexp_dir %s, epoch %d" % (self.exp_dir, epoch))
            batch_gen = self.corpus_reader.train_batch_gen()

            train_ler_total = 0
            batch_i = None
            print("\tBatch...", end="")
            for batch_i, batch in enumerate(batch_gen):
                print("%d..." % batch_i, end="")
                sys.stdout.flush()
                batch_x, batch_x_lens, batch_y = batch

                feed_dict = {self.batch_x: batch_x,
                             self.batch_x_lens: batch_x_lens,
                             self.batch_y: batch_y}

                _, ler, = sess.run([self.optimizer, self.ler],
                                   feed_dict=feed_dict)

                train_ler_total += ler

            feed_dict = {self.batch_x: valid_x,
                         self.batch_x_lens: valid_x_lens,
                         self.batch_y: valid_y}

            try:
                valid_ler, dense_decoded, dense_ref = sess.run(
                    [self.ler, self.dense_decoded, self.dense_ref],
                    feed_dict=feed_dict)
            except tf.errors.ResourceExhaustedError:
                print("Ran out of memory allocating a batch:")
                import pprint
                pprint.pprint(feed_dict)
                raise
            hyps, refs = self.corpus_reader.human_readable_hyp_ref(
                dense_decoded, dense_ref)
            # Log hypotheses
            with open(os.path.join(hyps_dir, "epoch%d_hyps" % epoch), "w") as hyps_f:
                for hyp in hyps:
                    print(" ".join(hyp), file=hyps_f)
            if epoch == 0:
                with open(os.path.join(hyps_dir, "refs"), "w") as refs_f:
                    for ref in refs:
                        print(" ".join(ref), file=refs_f)

            valid_per = utils.batch_per(hyps, refs)

            epoch_str = "Epoch %d. Training LER: %f, validation LER: %f, validation PER: %f" % (
                epoch, (train_ler_total / (batch_i + 1)), valid_ler, valid_per)
            print(epoch_str, flush=True, file=out_file)
            if best_epoch_str == None:
                best_epoch_str = epoch_str

            # Implement early stopping.
            if valid_ler < best_valid_ler:
                print("New best valid_ler", file=out_file)
                best_valid_ler = valid_ler
                best_epoch_str = epoch_str
                steps_since_last_record = 0

                # Save the model.
                path = os.path.join(self.exp_dir, "model", "model_best.ckpt")
                if not os.path.exists(os.path.dirname(path)):
                    os.mkdir(os.path.dirname(path))
                saver.save(sess, path)
                self.saved_model_path = path

                # Output best hyps
                with open(os.path.join(hyps_dir, "best_hyps"), "w") as hyps_f:
                    for hyp in hyps:
                        print(" ".join(hyp), file=hyps_f)

            else:
                print("Steps since last best valid_ler: %d" % (steps_since_last_record), file=out_file)
                steps_since_last_record += 1
                if epoch >= max_epochs:
                    with open(os.path.join(self.exp_dir, "best_scores.txt"), "w") as best_f:
                        print(best_epoch_str, file=best_f, flush=True)
                        sess.close()
                        out_file.close()
                        break
                if steps_since_last_record >= early_stopping_steps:
                    if epoch >= min_epochs:
                        # Then we've done the minimum number of epochs.
                        if valid_ler <= max_valid_ler and ler <= max_train_ler:
                            # Then training error has moved sufficiently
                            # towards convergence.
                            print("""Stopping since best validation score hasn't been
                                  beaten in %d epochs and at least %d have been
                                  done. The valid ler (%d) is below %d and 
                                  the train ler (%d) is below %d.""" %
                                  (early_stopping_steps, min_epochs, valid_ler,
                                  max_valid_ler, ler, max_train_ler),
                                  file=out_file, flush=True)
                            with open(os.path.join(self.exp_dir, "best_scores.txt"), "w") as best_f:
                                print(best_epoch_str, file=best_f, flush=True)
                                sess.close()
                                out_file.close()
                                break
                        else:
                            # Keep training because we haven't achieved
                            # convergence.
                            continue
                    else:
                        # Keep training because we haven't done the minimum
                        # numper of epochs.
                        continue

        # Finally, run evaluation on the test set.
        self.eval(restore_model_path=self.saved_model_path)

        sess.close()
        out_file.close()



# pylint: disable-all

""" Implement the CNN-based acoustic model of Zhang et al. 2017
    'Towards End-to-End Speech Recognition with Deep Convolutional Neural
    Networks'.
"""


import os
import random

import numpy as np
import tensorflow as tf

#from preprocess_timit import len_longest_timit_utterance
config = tf.ConfigProto()
config.gpu_options.allow_growth = True
sess = tf.InteractiveSession(config=config)

random.seed(0)

def count_parameters():
    total_parameters = 0
    for variable in tf.trainable_variables():
        # shape is an array of tf.Dimension
        shape = variable.get_shape()
        print(shape)
        print(len(shape))
        variable_parametes = 1
        for dim in shape:
            print(dim)
            variable_parametes *= dim.value
        print(variable_parametes)
        total_parameters += variable_parametes
    print(total_parameters)

def num_phones(path="/home/oadams/mam/data/timit/"):
    phn_set = set()
    for root, dirnames, filenames in os.walk(path):
        for fn in filenames:
            prefix = fn.split(".")[0] # Get the file prefix
            if fn.endswith(".phn"):
                path = os.path.join(root, fn)
                with open(path) as phn_f:
                    phns = phn_f.readline().split()
                for phn in phns:
                    phn_set.add(phn)
    return len(phn_set)

def load_timit(path="/home/oadams/mam/data/timit/train", rand=True,
        batch_size=10, labels="phonemes"):
    """ Load the already preprocessed TIMIT data. """

    def create_one_hot(i, num_classes):
        """ Takes a list or array with numbers representing classes, and
        creates a corresponding 2-dimensional array of one-hot vectors."""

        one_hot = np.zeros((1, num_classes))
        one_hot[0, i] = 1
        return one_hot

    train_paths = []
    dialect_classes = set()

    for root, dirnames, filenames in os.walk(path):
        for fn in filenames:
            prefix = fn.split(".")[0] # Get the file prefix
            if fn.endswith(".log_mel_filterbank.zero_pad.npy"):
                # Add to filename list.
                path = os.path.join(root, fn)
                train_paths.append(os.path.join(root, fn))
                dialect_classes.add(path.split("/")[-3])

    dialect_classes = sorted(list(dialect_classes))
    dr_class_map = dict(zip(dialect_classes, range(0, len(dialect_classes))))

    if rand:
        # Load a random subset of the training set of batch_size
        path_batch = random.sample(train_paths, batch_size)
        feats = [np.load(path) for path in path_batch]
        x = np.array(feats)
        if labels == "dialects":
            y_labels = np.array([dr_class_map[path.split("/")[-3]] for path in path_batch])
            # Make into one-hot vectors
            batch_y = np.concatenate([create_one_hot(label, len(dialect_classes))
                    for label in y_labels])
        elif labels == "phonemes":
            phn_paths = ["".join(path.split(".")[:-3])+".phn" for path in path_batch]
            batch_y = []
            for phn_path in phn_paths:
                with open(phn_path) as phn_f:
                    batch_y.append(phn_f.readline().split())
    else:
        raise NotImplementedError

    batch_x = np.array(feats)
    return batch_x, batch_y

# Create the graph
def weight_variable(shape):
    initial = tf.truncated_normal(shape, stddev=0.1)
    return tf.Variable(initial)

def bias_variable(shape):
    initial = tf.constant(0.1, shape=shape)
    return tf.Variable(initial)

def conv2d(x, W):
    # Shift the window by 1 in each direction
    return tf.nn.conv2d(x, W, strides=[1, 1, 1, 1], padding="SAME")

def max_pool_1x3(x):
    # Use max pooling over each 1x3 windows, as per Zhang et al.
    return tf.nn.max_pool(x, ksize=[1, 1, 3, 1], strides=[1, 1, 3, 1],
                          padding="SAME")


# There are 26 log Mel filterbank features and the 3 channels correspond to
# static, delta and double-delta.
x = tf.placeholder(tf.float32,
                    shape=[None, None, 26, 3])
# There are 8 dialect classes.
y = tf.placeholder(tf.float32, shape=[None, 8])

keep_prob = tf.placeholder(tf.float32)

# The 1st, 2nd and 4th dimensions I can arbitrarily pick, but the 3rd dimension
# is the input number of channels and must correspond to x. This weight becomes
# the filter argument to tf.nn.conv2d.
W_conv1 = weight_variable([5, 3, 3, 128])
b_conv1 = bias_variable([128])
h_conv1 = tf.nn.relu(conv2d(x, W_conv1) + b_conv1)
h_conv1_drop = tf.nn.dropout(h_conv1, keep_prob)

h_pool1 = max_pool_1x3(h_conv1_drop)
h_pool1_drop = tf.nn.dropout(h_pool1, keep_prob)

W_conv2 = weight_variable([5, 3, 128, 128])
b_conv2 = bias_variable([128])
h_conv2 = tf.nn.relu(conv2d(h_pool1_drop, W_conv2) + b_conv2)
h_conv2_drop = tf.nn.dropout(h_conv2, keep_prob)

W_conv3 = weight_variable([5, 3, 128, 128])
b_conv3 = bias_variable([128])
h_conv3 = tf.nn.relu(conv2d(h_conv2_drop, W_conv3) + b_conv3)
h_conv3_drop = tf.nn.dropout(h_conv3, keep_prob)

W_conv4 = weight_variable([5, 3, 128, 128])
b_conv4 = bias_variable([128])
h_conv4 = tf.nn.relu(conv2d(h_conv3_drop, W_conv4) + b_conv4)
h_conv4_drop = tf.nn.dropout(h_conv4, keep_prob)

W_conv5 = weight_variable([5, 3, 128, 256])
b_conv5 = bias_variable([256])
h_conv5 = tf.nn.relu(conv2d(h_conv4_drop, W_conv5) + b_conv5)
h_conv5_drop = tf.nn.dropout(h_conv5, keep_prob)

W_conv6 = weight_variable([5, 3, 256, 256])
b_conv6 = bias_variable([256])
h_conv6 = tf.nn.relu(conv2d(h_conv5_drop, W_conv6) + b_conv6)
h_conv6_drop = tf.nn.dropout(h_conv6, keep_prob)

W_conv7 = weight_variable([5, 3, 256, 256])
b_conv7 = bias_variable([256])
h_conv7 = tf.nn.relu(conv2d(h_conv6_drop, W_conv7) + b_conv7)
h_conv7_drop = tf.nn.dropout(h_conv7, keep_prob)

W_conv8 = weight_variable([5, 3, 256, 256])
b_conv8 = bias_variable([256])
h_conv8 = tf.nn.relu(conv2d(h_conv7_drop, W_conv8) + b_conv8)
h_conv8_drop = tf.nn.dropout(h_conv8, keep_prob)

W_conv9 = weight_variable([5, 3, 256, 256])
b_conv9 = bias_variable([256])
h_conv9 = tf.nn.relu(conv2d(h_conv8_drop, W_conv9) + b_conv9)
h_conv9_drop = tf.nn.dropout(h_conv9, keep_prob)

W_conv10 = weight_variable([5, 3, 256, 256])
b_conv10 = bias_variable([256])
h_conv10 = tf.nn.relu(conv2d(h_conv9_drop, W_conv10) + b_conv10)

h_conv10_flat = tf.reshape(h_conv10, [-1, 778*9*256])
h_conv10_flat_drop = tf.nn.dropout(h_conv10_flat, keep_prob)

W_fc1 = weight_variable([778*9*256, 128])
b_fc1 = bias_variable([128])
h_fc1 = tf.nn.relu(tf.matmul(h_conv10_flat_drop, W_fc1) + b_fc1)
h_fc1_drop = tf.nn.dropout(h_fc1, keep_prob)

W_fc2 = weight_variable([128, 128])
b_fc2 = bias_variable([128])
h_fc2 = tf.nn.relu(tf.matmul(h_fc1_drop, W_fc2) + b_fc2)
h_fc2_drop = tf.nn.dropout(h_fc2, keep_prob)

num_labels = num_phones() + 1

W_fc3 = weight_variable([128, num_labels])
b_fc3 = bias_variable([num_labels])
y_fc = tf.matmul(h_fc2_drop, W_fc3) + b_fc3

cross_entropy = tf.reduce_mean(
        tf.nn.softmax_cross_entropy_with_logits(labels=y, logits=y_fc))
train_step = tf.train.GradientDescentOptimizer(0.5).minimize(cross_entropy)

correct_prediction = tf.equal(tf.argmax(y_fc, 1), tf.argmax(y, 1))
accuracy = tf.reduce_mean(tf.cast(correct_prediction, tf.float32))

tf.initialize_all_variables().run()

total_acc = 0.0
for i in range(1000):
    batch = load_timit(batch_size=10)
    train_accuracy = accuracy.eval(
            feed_dict={x:batch[0], y: batch[1], keep_prob: 0.7})
    total_acc += train_accuracy
    print("Step %d, training accuracy %g, avg acc: %g" % (i, train_accuracy, total_acc/i))
    train_step.run(
            feed_dict={x:batch[0], y: batch[1], keep_prob: 0.7})



""" An interface with the Na data. """

import os
import random
import subprocess
from subprocess import PIPE

import numpy as np
import xml.etree.ElementTree as ET

from .. import config
from .. import corpus
from .. import feat_extract
from .. import utils
from ..exceptions import PersephoneException
from . import pangloss

ORG_DIR = config.NA_DIR
# TODO eventually remove "new" when ALTA experiments are finished.
TGT_DIR = os.path.join(config.TGT_DIR, "na", "new")
ORG_XML_DIR = os.path.join(ORG_DIR, "xml")
ORG_WAV_DIR = os.path.join(ORG_DIR, "wav")
TGT_WAV_DIR = os.path.join(TGT_DIR, "wav")
FEAT_DIR = os.path.join(TGT_DIR, "feat")
LABEL_DIR = os.path.join(TGT_DIR, "label")
TRANSL_DIR = os.path.join(TGT_DIR, "transl")

# The directory for untranscribed audio we want to transcribe with automatic
# methods.
UNTRAN_DIR = os.path.join(TGT_DIR, "untranscribed")

#PREFIXES = [os.path.splitext(fn)[0]
#            for fn in os.listdir(ORG_TRANSCRIPT_DIR)
#            if fn.endswith(".txt")]

# TODO Move into feat creation functions.
if not os.path.isdir(TGT_DIR):
    os.makedirs(TGT_DIR)

if not os.path.isdir(FEAT_DIR):
    os.makedirs(FEAT_DIR)

# HARDCODED values
MISC_SYMBOLS = [' ̩', '~', '=', ':', 'F', '¨', '↑', '“', '”', '…', '«', '»',
'D', 'a', 'ː', '#', '$', "‡"]
BAD_NA_SYMBOLS = ['D', 'F', '~', '…', '=', '↑', ':']
PUNC_SYMBOLS = [',', '!', '.', ';', '?', "'", '"', '*', ':', '«', '»', '“', '”', "ʔ"]
UNI_PHNS = {'q', 'p', 'ɭ', 'ɳ', 'h', 'ʐ', 'n', 'o', 'ɤ', 'ʝ', 'ɛ', 'g',
            'i', 'u', 'b', 'ɔ', 'ɯ', 'v', 'ɑ', 'l', 'ɖ', 'ɻ', 'ĩ', 'm',
            't', 'w', 'õ', 'ẽ', 'd', 'ɣ', 'ɕ', 'c', 'ʁ', 'ʑ', 'ʈ', 'ɲ', 'ɬ',
            's', 'ŋ', 'ə', 'e', 'æ', 'f', 'j', 'k', 'z', 'ʂ'}
BI_PHNS = {'dʑ', 'ẽ', 'ɖʐ', 'w̃', 'æ̃', 'qʰ', 'i͂', 'tɕ', 'v̩', 'o̥', 'ts',
           'ɻ̩', 'ã', 'ə̃', 'ṽ', 'pʰ', 'tʰ', 'ɤ̃', 'ʈʰ', 'ʈʂ', 'ɑ̃', 'ɻ̃', 'kʰ',
           'ĩ', 'õ', 'dz', "ɻ̍", "wæ", "wɑ", "wɤ", "jæ", "jɤ", "jo"}
FILLERS = {"əəə…", "mmm…"}
TRI_PHNS = {"tɕʰ", "ʈʂʰ", "tsʰ", "ṽ̩", "ṽ̩", "ɻ̩̃", "wæ̃", "w̃æ"}
UNI_TONES = {"˩", "˥", "˧"}
BI_TONES = {"˧˥", "˩˥", "˩˧", "˧˩"}
TONES = UNI_TONES.union(BI_TONES)
SYMBOLS_TO_PREDICT = {"|"}

PHONEMES = UNI_PHNS.union(BI_PHNS).union(TRI_PHNS).union(FILLERS)

# TODO Get rid of these variables, as they're not used in the class, only for
# preparing phonemes_onehot feats.
PHONEMES_TO_INDICES = {phn: index for index, phn in enumerate(PHONEMES)}
INDICES_TO_PHONEMES = {index: phn for index, phn in enumerate(PHONEMES)}

# TODO Potentially remove?
#PHONES_TONES = sorted(list(PHONES.union(set(TONES)))) # Sort for determinism
#PHONESTONES2INDICES = {phn_tone: index for index, phn_tone in enumerate(PHONES_TONES)}
#INDICES2PHONESTONES = {index: phn_tone for index, phn_tone in enumerate(PHONES_TONES)}
#TONES2INDICES = {tone: index for index, tone in enumerate(TONES)}
#INDICES2TONES = {index: tone for index, tone in enumerate(TONES)}

def preprocess_na(sent, label_type):

    if label_type == "phonemes_and_tones":
        phonemes = True
        tones = True
        tgm = True
    elif label_type == "phonemes_and_tones_no_tgm":
        phonemes = True
        tones = True
        tgm = False
    elif label_type == "phonemes":
        phonemes = True
        tones = False
    elif label_type == "tones":
        phonemes = False
        tones = True
        tgm = True
    elif label_type == "tones_notgm":
        phonemes = False
        tones = True
        tgm = False
    else:
        raise ValueError("Unrecognized label type: %s" % label_type)

    def pop_phoneme(sentence):
        # TODO desperately needs refactoring

        # Treating fillers as single tokens; normalizing to əəə and mmm
        if phonemes:
            if sentence[:4] in ["əəə…", "mmm…"]:
                return sentence[:4], sentence[4:]
            if sentence.startswith("ə…"):
                return "əəə…", sentence[2:]
            if sentence.startswith("m…"):
                return "mmm…", sentence[2:]
            if sentence.startswith("mm…"):
                return "mmm…", sentence[3:]

        # Normalizing some stuff
        if sentence[:3] == "wæ̃":
            if phonemes:
                return "w̃æ", sentence[3:]
            else:
                return None, sentence[3:]
        if sentence[:3] == "ṽ̩":
            if phonemes:
                return "ṽ̩", sentence[3:]
            else:
                return None, sentence[3:]

        if sentence[:3] in TRI_PHNS:
            if phonemes:
                return sentence[:3], sentence[3:]
            else:
                return None, sentence[3:]
        if sentence[:2] in BI_PHNS:
            if phonemes:
                return sentence[:2], sentence[2:]
            else:
                return None, sentence[2:]
        if sentence[0] in UNI_PHNS:
            if phonemes:
                return sentence[0], sentence[1:]
            else:
                return None, sentence[1:]
        if sentence[:2] in BI_TONES:
            if tones:
                return sentence[:2], sentence[2:]
            else:
                return None, sentence[2:]
        if sentence[0] in UNI_TONES:
            if tones:
                return sentence[0], sentence[1:]
            else:
                return None, sentence[1:]
        if sentence[0] in MISC_SYMBOLS:
            # We assume these symbols cannot be captured.
            return None, sentence[1:]
        if sentence[0] in BAD_NA_SYMBOLS:
            return None, sentence[1:]
        if sentence[0] in PUNC_SYMBOLS:
            return None, sentence[1:]
        if sentence[0] in ["-", "ʰ", "/"]:
            return None, sentence[1:]
        if sentence[0] in set(["<", ">"]):
            # We keep everything literal, thus including what is in <>
            # brackets; so we just remove these tokens"
            return None, sentence[1:]
        if sentence[0] == "[":
            # It's an opening square bracket, so ignore everything until we
            # find a closing one.
            if sentence.find("]") == len(sentence)-1:
                # If the closing bracket is the last char
                return None, ""
            else:
                return None, sentence[sentence.find("]")+1:]
        if sentence[0] in set([" ", "\t", "\n"]):
            # Return a space char so that it can be identified in word segmentation
            # processing.
            return " ", sentence[1:]
        if sentence[0] == "|" or sentence[0] == "ǀ" or sentence[0] == "◊":
            # TODO Address extrametrical span symbol ◊ differently. For now,
            # treating it as a tone group boundary marker for consistency with
            # previous work.
            if tgm:
                return "|", sentence[1:]
            else:
                return None, sentence[1:]
        if sentence[0] in "()":
            return None, sentence[1:]
        print("***" + sentence)
        raise ValueError("Next character not recognized: " + sentence[:1])

    def filter_for_phonemes(sentence):
        """ Returns a sequence of phonemes and pipes (word delimiters). Tones,
        syllable boundaries, whitespace are all removed."""

        filtered_sentence = []
        while sentence != "":
            phoneme, sentence = pop_phoneme(sentence)
            if phoneme != " ":
                filtered_sentence.append(phoneme)
        filtered_sentence = [item for item in filtered_sentence if item != None]
        return " ".join(filtered_sentence)

    # Filter utterances with certain words
    if "BEGAIEMENT" in sent:
        return ""
    sent = filter_for_phonemes(sent)
    return sent

def preprocess_french(trans, fr_nlp, remove_brackets_content=True):
    """ Takes a list of sentences in french and preprocesses them."""

    if remove_brackets_content:
        trans = pangloss.remove_content_in_brackets(trans, "[]")
    # Not sure why I have to split and rejoin, but that fixes a Spacy token
    # error.
    trans = fr_nlp(" ".join(trans.split()[:]))
    #trans = fr_nlp(trans)
    trans = " ".join([token.lower_ for token in trans if not token.is_punct])

    return trans

def trim_wavs(org_wav_dir=ORG_WAV_DIR,
              tgt_wav_dir=TGT_WAV_DIR,
              org_xml_dir=ORG_XML_DIR):
    """ Extracts sentence-level transcriptions, translations and wavs from the
    Na Pangloss XML and WAV files. But otherwise doesn't preprocess them."""

    print("Trimming wavs...")

    if not os.path.exists(os.path.join(tgt_wav_dir, "TEXT")):
        os.makedirs(os.path.join(tgt_wav_dir, "TEXT"))
    if not os.path.exists(os.path.join(tgt_wav_dir, "WORDLIST")):
        os.makedirs(os.path.join(tgt_wav_dir, "WORDLIST"))

    for fn in os.listdir(org_xml_dir):
        print(fn)
        path = os.path.join(org_xml_dir, fn)
        prefix, _ = os.path.splitext(fn)

        if os.path.isdir(path):
            continue
        if not path.endswith(".xml"):
            continue

        rec_type, sents, times, transls = pangloss.get_sents_times_and_translations(path)

        # Extract the wavs given the times.
        for i, (start_time, end_time) in enumerate(times):
            if prefix.endswith("PLUSEGG"):
                in_wav_path = os.path.join(org_wav_dir, prefix.upper()[:-len("PLUSEGG")]) + ".wav"
            else:
                in_wav_path = os.path.join(org_wav_dir, prefix.upper()) + ".wav"
            headmic_path = os.path.join(org_wav_dir, prefix.upper()) + "_HEADMIC.wav"
            if os.path.isfile(headmic_path):
                in_wav_path = headmic_path

            out_wav_path = os.path.join(tgt_wav_dir, rec_type, "%s.%d.wav" % (prefix, i))
            assert os.path.isfile(in_wav_path)
            utils.trim_wav(in_wav_path, out_wav_path, start_time, end_time)

# Commenting out because of spacy requirement and working with the French
# translations isn't very integral to persephone.
#def prepare_transls():
#    """ Prepares the French translations. """
#
#    import spacy
#    fr_nlp = spacy.load("fr")
#
#    if not os.path.exists(os.path.join(TRANSL_DIR, "TEXT")):
#        os.makedirs(os.path.join(TRANSL_DIR, "TEXT"))
#    if not os.path.exists(os.path.join(TRANSL_DIR, "WORDLIST")):
#        os.makedirs(os.path.join(TRANSL_DIR, "WORDLIST"))
#
#    for fn in os.listdir(ORG_XML_DIR):
#        print(fn)
#        path = os.path.join(ORG_XML_DIR, fn)
#        prefix, _ = os.path.splitext(fn)
#
#        rec_type, sents, times, transls = pangloss.get_sents_times_and_translations(path)
#
#        # Tokenize the French translations and write them to file.
#        transls = [preprocess_french(transl[0], fr_nlp) for transl in transls]
#        for i, transl in enumerate(transls):
#            out_prefix = "%s.%d" % (prefix, i)
#            transl_path = os.path.join(TRANSL_DIR, rec_type, out_prefix + ".fr.txt")
#            with open(transl_path, "w") as transl_f:
#                print(transl, file=transl_f)

def prepare_labels(label_type, org_xml_dir=ORG_XML_DIR, label_dir=LABEL_DIR):
    """ Prepare the neural network output targets."""

    if not os.path.exists(os.path.join(label_dir, "TEXT")):
        os.makedirs(os.path.join(label_dir, "TEXT"))
    if not os.path.exists(os.path.join(label_dir, "WORDLIST")):
        os.makedirs(os.path.join(label_dir, "WORDLIST"))

    for fn in os.listdir(org_xml_dir):
        print(fn)

        path = os.path.join(org_xml_dir, fn)
        prefix, _ = os.path.splitext(fn)
        if os.path.isdir(path):
            continue
        if not path.endswith(".xml"):
            continue

        rec_type, sents, times, transls = pangloss.get_sents_times_and_translations(path)
        # Write the sentence transcriptions to file
        sents = [preprocess_na(sent, label_type) for sent in sents]
        for i, sent in enumerate(sents):
            if sent.strip() == "":
                # Then there's no transcription, so ignore this.
                continue
            out_fn = "%s.%d.%s" % (prefix, i, label_type)
            sent_path = os.path.join(label_dir, rec_type, out_fn)
            with open(sent_path, "w") as sent_f:
                print(sent, file=sent_f)

# TODO Consider factoring out as non-Na specific.
def prepare_untran(feat_type="fbank_and_pitch"):
    """ Preprocesses untranscribed audio."""
    org_dir = os.path.join(UNTRAN_DIR, "org")
    wav_dir = os.path.join(UNTRAN_DIR, "wav")
    feat_dir = os.path.join(UNTRAN_DIR, "feat")
    if not os.path.isdir(wav_dir):
        os.makedirs(wav_dir)
    if not os.path.isdir(feat_dir):
        os.makedirs(feat_dir)

    # Standardize into wav files.
    for fn in os.listdir(org_dir):
        in_path = os.path.join(org_dir, fn)
        prefix, _ = os.path.splitext(fn)
        mono16k_wav_path = os.path.join(wav_dir, "%s.wav" % prefix)
        if not os.path.isfile(mono16k_wav_path):
            feat_extract.convert_wav(in_path, mono16k_wav_path)

    # Split up the wavs
    wav_fns = os.listdir(wav_dir)
    for fn in wav_fns:
        in_fn = os.path.join(wav_dir, fn)
        prefix, _ = os.path.splitext(fn)
        # Split into sub-wavs and perform feat extraction.
        split_id = 0
        start, end = 0, 10 #in seconds
        length = utils.wav_length(in_fn)
        while True:
            out_fn = os.path.join(feat_dir, "%s.%d.wav" % (prefix, split_id))
            utils.trim_wav(in_fn, out_fn, start, end)
            if end > length:
                break
            start += 10
            end += 10
            split_id += 1

    # Do feat extraction.
    feat_extract.from_dir(os.path.join(feat_dir), feat_type=feat_type)

# TODO Consider factoring out as non-Na specific
def prepare_feats(feat_type, org_wav_dir=ORG_WAV_DIR, feat_dir=FEAT_DIR, tgt_wav_dir=TGT_WAV_DIR,
                  org_xml_dir=ORG_XML_DIR, label_dir=LABEL_DIR):
    """ Prepare the input features."""

    if not os.path.isdir(os.path.join(feat_dir, "WORDLIST")):
        os.makedirs(os.path.join(feat_dir, "WORDLIST"))
    if not os.path.isdir(os.path.join(feat_dir, "TEXT")):
        os.makedirs(os.path.join(feat_dir, "TEXT"))

    # Extract utterances from WAVS.
    trim_wavs(org_wav_dir=org_wav_dir,
              tgt_wav_dir=tgt_wav_dir,
              org_xml_dir=org_xml_dir)

    # TODO Currently assumes that the wav trimming from XML has already been
    # done.
    prefixes = []
    for fn in os.listdir(os.path.join(tgt_wav_dir, "WORDLIST")):
        if fn.endswith(".wav"):
            pre, _ = os.path.splitext(fn)
            prefixes.append(os.path.join("WORDLIST", pre))
    for fn in os.listdir(os.path.join(tgt_wav_dir, "TEXT")):
        if fn.endswith(".wav"):
            pre, _ = os.path.splitext(fn)
            prefixes.append(os.path.join("TEXT", pre))

    if feat_type=="phonemes_onehot":
        import numpy as np
        #prepare_labels("phonemes")
        for prefix in prefixes:
            label_fn = os.path.join(label_dir, "%s.phonemes" % prefix)
            out_fn = os.path.join(feat_dir, "%s.phonemes_onehot" %  prefix)
            try:
                with open(label_fn) as label_f:
                    labels = label_f.readlines()[0].split()
            except FileNotFoundError:
                continue
            indices = [PHONEMES_TO_INDICES[label] for label in labels]
            one_hots = one_hots = [[0]*len(PHONEMES) for _ in labels]
            for i, index in enumerate(indices):
                one_hots[i][index] = 1
                one_hots = np.array(one_hots)
                np.save(out_fn, one_hots)
    else:
        # Otherwise, 
        for prefix in prefixes:
            # Convert the wave to 16k mono.
            wav_fn = os.path.join(tgt_wav_dir, "%s.wav" % prefix)
            mono16k_wav_fn = os.path.join(feat_dir, "%s.wav" % prefix)
            if not os.path.isfile(mono16k_wav_fn):
                feat_extract.convert_wav(wav_fn, mono16k_wav_fn)

        # Extract features from the wavs.
        feat_extract.from_dir(os.path.join(feat_dir, "WORDLIST"), feat_type=feat_type)
        feat_extract.from_dir(os.path.join(feat_dir, "TEXT"), feat_type=feat_type)

def get_story_prefixes(label_type, label_dir=LABEL_DIR):
    """ Gets the Na text prefixes. """
    prefixes = [prefix for prefix in os.listdir(os.path.join(label_dir, "TEXT"))
                if prefix.endswith(".%s" % label_type)]
    prefixes = [os.path.splitext(os.path.join("TEXT", prefix))[0]
                for prefix in prefixes]
    return prefixes

def make_data_splits(label_type, train_rec_type="text_and_wordlist", max_samples=1000,
                     seed=0, tgt_dir=TGT_DIR):
    """ Creates a file with a list of prefixes (identifiers) of utterances to
    include in the test set. Test utterances must never be wordlists. Assumes
    preprocessing of label dir has already been done."""

    feat_dir = os.path.join(tgt_dir, "feat")
    test_prefix_fn=os.path.join(tgt_dir, "test_prefixes.txt")
    valid_prefix_fn=os.path.join(tgt_dir, "valid_prefixes.txt")
    with open(test_prefix_fn) as f:
        prefixes = f.readlines()
        test_prefixes = [("TEXT/" + prefix).strip() for prefix in prefixes]
    with open(valid_prefix_fn) as f:
        prefixes = f.readlines()
        valid_prefixes = [("TEXT/" + prefix).strip() for prefix in prefixes]

    label_dir = os.path.join(tgt_dir, "label")
    prefixes = get_story_prefixes(label_type, label_dir=label_dir)
    prefixes = list(set(prefixes) - set(valid_prefixes))
    prefixes = list(set(prefixes) - set(test_prefixes))
    prefixes = utils.filter_by_size(
        feat_dir, prefixes, "fbank", max_samples)

    if train_rec_type == "text":
        train_prefixes = prefixes
    else:
        wordlist_prefixes = [prefix for prefix in os.listdir(os.path.join(label_dir, "WORDLIST"))
                             if prefix.endswith("phonemes")]
        wordlist_prefixes = [os.path.splitext(os.path.join("WORDLIST", prefix))[0]
                             for prefix in wordlist_prefixes]
        wordlist_prefixes = utils.filter_by_size(
                feat_dir, wordlist_prefixes, "fbank", max_samples)
        if train_rec_type == "wordlist":
            prefixes = wordlist_prefixes
        elif train_rec_type == "text_and_wordlist":
            prefixes.extend(wordlist_prefixes)
        else:
            raise PersephoneException("train_rec_type='%s' not supported." % train_rec_type)
        train_prefixes = prefixes
    random.seed(0)
    random.shuffle(train_prefixes)

    return train_prefixes, valid_prefixes, test_prefixes

def get_stories(label_type):
    """ Returns a list of the stories in the Na corpus. """

    prefixes = get_story_prefixes(label_type)
    texts = list(set([prefix.split(".")[0].split("/")[1] for prefix in prefixes]))
    return texts

def make_story_splits(valid_story, test_story, max_samples, label_type, tgt_dir=TGT_DIR):

    feat_dir=os.path.join(tgt_dir, "feat/")

    prefixes = get_story_prefixes(label_type)
    # TODO Remove assumption of fbank features
    prefixes = utils.filter_by_size(
        feat_dir, prefixes, "fbank", max_samples)

    train = []
    valid = []
    test = []
    for prefix in prefixes:
        if valid_story == os.path.basename(prefix).split(".")[0]:
            valid.append(prefix)
        elif test_story == os.path.basename(prefix).split(".")[0]:
            test.append(prefix)
        else:
            train.append(prefix)

    # Sort by utterance integer value
    test.sort(key=lambda x: int(x.split(".")[-1]))
    valid.sort(key=lambda x: int(x.split(".")[-1]))

    return train, valid, test

class Corpus(corpus.AbstractCorpus):
    """ Class to interface with the Na corpus. """

    def __init__(self,
                 feat_type="fbank_and_pitch",
                 label_type="phonemes_and_tones",
                 train_rec_type="text", max_samples=1000,
                 valid_story=None, test_story=None,
                 tgt_dir=TGT_DIR):
        super().__init__(feat_type, label_type)

        self.FEAT_DIR = os.path.join(tgt_dir, "feat")
        self.LABEL_DIR = os.path.join(tgt_dir, "label")
        self.UNTRAN_FEAT_DIR = os.path.join(UNTRAN_DIR, "feat")

        self.max_samples = max_samples
        self.train_rec_type = train_rec_type

        if label_type == "phonemes_and_tones":
            self.labels = PHONEMES.union(TONES).union(SYMBOLS_TO_PREDICT)
        elif label_type == "phonemes_and_tones_no_tgm":
            self.labels = PHONEMES.union(TONES)
        elif label_type == "phonemes":
            self.labels = PHONEMES
        elif label_type == "tones":
            self.labels = TONES.union(SYMBOLS_TO_PREDICT)
        elif label_type == "tones_notgm":
            self.labels = TONES
        else:
            raise PersephoneException("label_type %s not implemented." % label_type)

        self.feat_type = feat_type
        self.label_type = label_type

        self.valid_story = valid_story
        self.test_story = test_story

        # TODO Make this also work with wordlists.
        if valid_story or test_story:
            if not (valid_story and test_story):
                raise PersephoneException(
                    "We need a valid story if we specify a test story "
                    "and vice versa. This shouldn't be required but for "
                    "now it is.")

            train, valid, test = make_story_splits(valid_story, test_story,
                                                   max_samples,
                                                   self.label_type,
                                                   tgt_dir=tgt_dir)
        else:
            train, valid, test = make_data_splits(self.label_type,
                                                  train_rec_type=train_rec_type,
                                                  max_samples=max_samples,
                                                  tgt_dir=tgt_dir)
        self.train_prefixes = train
        self.valid_prefixes = valid
        self.test_prefixes = test

        self.LABEL_TO_INDEX = {label: index for index, label in enumerate(
                                 ["pad"] + sorted(list(self.labels)))}
        self.INDEX_TO_LABEL = {index: phn for index, phn in enumerate(
                                 ["pad"] + sorted(list(self.labels)))}
        self.vocab_size = len(self.labels)

    def output_story_prefixes(self):
        """ Writes the set of prefixes to a file this is useful for pretty
        printing in results.latex_output. """

        if not self.test_story:
            raise NotImplementedError(
                "I want to write the prefixes to a file"
                "called <test_story>_prefixes.txt, but there's no test_story.")

        fn = os.path.join(TGT_DIR, "%s_prefixes.txt" % self.test_story)
        with open(fn, "w") as f:
            for utter_id in self.test_prefixes:
                print(utter_id.split("/")[1], file=f)

    # TODO Use 'labels' instead of 'phonemes' here and in corpus.py
    # Also, factor out as non-Na-specific.
    def indices_to_phonemes(self, indices):
        return [(self.INDEX_TO_LABEL[index]) for index in indices]
    def phonemes_to_indices(self, labels):
        return [self.LABEL_TO_INDEX[label] for label in labels]

    def __repr__(self):
        return ("%s(" % self.__class__.__name__ +
                "feat_type=\"%s\",\n" % self.feat_type +
                "\tlabel_type=\"%s\",\n" % self.label_type +
                "\ttrain_rec_type=\"%s\",\n" % self.train_rec_type +
                "\tmax_samples=%s,\n" % self.max_samples +
                "\tvalid_story=%s,\n" % repr(self.valid_story) +
                "\ttest_story=%s)\n" % repr(self.test_story))




import numpy as np

def min_edit_distance(source, target,
        ins_cost=lambda _x: 1,
        del_cost=lambda _x: 1,
        sub_cost=lambda x, y: 0 if x == y else 1):
    """Finds the minimum edit distance between two strings using the
    Levenshtein weighting as a default, but offers keyword arguments to supply
    functions to measure the costs for editing with different characters.

    ins_cost -- A function describing the cost of inserting a given char
    del_cost -- A function describing the cost of deleting a given char
    sub_cost -- A function describing the cost of substituting one char for
    another

    """

    # Initialize an m+1 by n+1 array. Note that the strings start from index 1,
    # with index 0 being used to denote the empty string.
    n = len(target)
    m = len(source)
    distance = np.zeros([m+1, n+1], dtype=np.int16)

    # Initialize the zeroth row and column to be the distance from the empty
    # string.
    for i in range(1, m+1):
        distance[i, 0] = distance[i-1, 0] + ins_cost(source[i-1])
    for j in range(1, n+1):
        distance[0, j] = distance[0, j-1] + ins_cost(target[j-1])

    # Do the dynamic programming to fill in the matrix with the edit distances.
    for j in range(1, n+1):
        for i in range(1, m+1):
            distance[i, j] = min(
                    distance[i-1, j] + ins_cost(source[i-1]),
                    distance[i-1, j-1] + sub_cost(source[i-1],target[j-1]),
                    distance[i, j-1] + del_cost(target[j-1]))

    return distance[len(source), len(target)]

def print_matrix(matrix):
    for row in matrix:
        print(row)

def min_edit_distance_align(source, target,
        ins_cost=lambda _x: 1,
        del_cost=lambda _x: 1,
        sub_cost=lambda x, y: 0 if x == y else 1):
    """Finds a minimum cost alignment between two strings using the
    Levenshtein weighting as a default, but offering keyword arguments to
    supply functions to measure the costs for editing with different
    characters.

    ins_cost -- A function describing the cost of inserting a given char
    del_cost -- A function describing the cost of deleting a given char
    sub_cost -- A function describing the cost of substituting one char for
    another

    """

    # Initialize an m+1 by n+1 array to hold the distances, and an equal sized
    # array to store the backpointers. Note that the strings start from index
    # 1, with index 0 being used to denote the empty string.
    n = len(target)
    m = len(source)
    dist = [[0]*(n+1) for _ in range(m+1)]
    bptrs = [[[]]*(n+1) for _ in range(m+1)]

    # Adjust the initialization of the first column and row of the matrices to
    # their appropriate values.
    for i in range(1, m+1):
        dist[i][0] = i
        bptrs[i][0] = (i-1, 0)
    for j in range(1, n+1):
        dist[0][j] = j
        bptrs[0][j] = (0, j-1)

    # Do the dynamic programming to fill in the matrix with the edit distances.
    for j in range(1, n+1):
        for i in range(1, m+1):
            options = [
                    (dist[i-1][j] + ins_cost(target[j-1]),
                        (i-1, j)),
                    (dist[i-1][j-1] + sub_cost(source[i-1],target[j-1]),
                        (i-1, j-1)),
                    (dist[i][j-1] + del_cost(source[i-1]),
                        (i, j-1))]
            (minimum, pointer) = sorted(options)[0]
            dist[i][j] = minimum
            bptrs[i][j] = pointer

    # Put the backtrace in a list, and reverse it to get a forward trace.
    bt = [(m,n)]
    cell = bptrs[m][n]
    while True:
        bt.append(cell)
        if bptrs[cell[0]][cell[1]]:
            cell = bptrs[cell[0]][cell[1]]
        else:
            break
    trace = list(reversed(bt))

    # Construct an alignment between source and target using the trace.
    alignment = []
    for i in range(1, len(trace)):
        current = trace[i]
        prev = trace[i-1]

        # If the cell is diagonal from the previous one, it's a substitution or
        # there wasn't a change.
        if current[0] - prev[0] == 1 and current[1] - prev[1] == 1:
            alignment.append((source[current[0]-1], target[current[1]-1]))
        # Otherwise if it moves only on the source side, it's a deletion
        elif current[0] - prev[0] == 1:
            alignment.append((source[current[0]-1], ""))
        # Otherwise if it moves only on the target side, it's an insertion
        elif current[1] - prev[1] == 1:
            alignment.append(("", target[current[1]-1]))

    return alignment

def cluster_alignment_errors(alignment):
    """Takes an alignment created by min_edit_distance_align() and groups
    consecutive errors together. This is useful, because there are often
    many possible alignments, and so often we can't meaningfully distinguish
    between alignment errors at the character level, so it makes many-to-many
    mistakes more readable."""

    newalign = []
    mistakes = ([],[])
    for align_item in alignment:
        if align_item[0] == align_item[1]:
            if mistakes != ([],[]):
                newalign.append((tuple(mistakes[0]), tuple(mistakes[1])))
                mistakes = ([],[])
            newalign.append((tuple([align_item[0]]), tuple([align_item[1]])))
        else:
            if align_item[0] != "":
                mistakes[0].append(align_item[0])
            if align_item[1] != "":
                mistakes[1].append(align_item[1])
    if mistakes != ([],[]):
        newalign.append((tuple(mistakes[0]), tuple(mistakes[1])))
        mistakes = ([],[])

    return newalign

def word_error_rate(ref, hypo):
    """Returns the word error rate of the supplied hypothesis with respect to
    the reference string."""

    distance = min_edit_distance(ref, hypo)
    return 100 * float(distance) / len(ref)



"""Context manager for changing the current working directory"""

import os

class cd:
    """Context manager for changing the current working directory"""
    def __init__(self, newPath):
        self.newPath = os.path.expanduser(newPath)

    def __enter__(self):
        self.savedPath = os.getcwd()
        os.chdir(self.newPath)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.savedPath)



"""
Offers functions for tokenizing utterances into phonemes, characters or
other symbols.
"""

from typing import Iterable

def segment_into_chars(utterance: str) -> str:
    """ Segments an utterance into space delimited characters. """

    if not isinstance(utterance, str):
        raise TypeError("Input type must be a string. Got {}.".format(type(utterance)))

    utterance.strip()
    utterance = utterance.replace(" ", "")
    return " ".join(utterance)

def segment_into_tokens(utterance: str, token_inventory: Iterable[str]):
    """
    Segments an utterance (a string) into tokens based on an inventory of
    tokens (a list or set of strings).

    The approach: Given the rest of the utterance, find the largest token (in
    character length) that is found in the token_inventory, and treat that as a
    token before segmenting the rest of the string.

    Note: Your orthography may open the door to ambiguities in the
    segmentation. Hopefully not, but another alternative it to simply segment
    on characters with segment_into_chars()
    """

    if not isinstance(utterance, str):
        raise TypeError("Input type must be a string. Got {}.".format(type(utterance)))

    # Token inventory needs to be hashable for speed
    token_inventory = set(token_inventory)
    # Get the size of the longest token in the inventory
    max_len = len(sorted(list(token_inventory), key=lambda x: len(x))[-1])

    def segment_token(utterance):
        if utterance == "":
            return "", ""

        for i in range(max_len, 0, -1):
            if utterance[:i] in token_inventory:
                return utterance[:i], utterance[i:]
        # If the next character is preventing segmentation, move on.
        return "", utterance[1:]

    tokens = []
    head, tail = segment_token(utterance)
    tokens.append(head)
    while tail != "":
        head, tail = segment_token(tail)
        tokens.append(head)
    tokens = [tok for tok in tokens if tok != ""]

    return " ".join(tokens)




""" Describes the abstract Corpus class that all interfaces to corpora should
    subclass. """

import abc
from collections import namedtuple
import os
from os.path import join
import random

import numpy as np

from . import feat_extract
from . import utils
from .exceptions import PersephoneException

class AbstractCorpus(metaclass=abc.ABCMeta):
    "All interfaces to corpora are subclasses of this class."""

    def get_target_prefix(self, prefix):
        """ Gets the target gfn given a prefix. """

        prefix = os.path.basename(prefix)
        return os.path.join(self.TGT_DIR, "transcriptions", "utterances", prefix)

    def __init__(self, feat_type, target_type, max_samples=None):
        """ feat_type: A string describing the input features. For
                       example, 'log_mel_filterbank'.
            target_type: A string describing the targets. For example,
                         'phonemes' or 'tones'.
            max_samples: The maximum length in samples of a train, valid, or
                         test utterance. Used to save memory in exchange for
                         reducing the number of effective training examples.
        """
        self.feat_type = feat_type
        self.target_type = target_type
        self.label_type = target_type
        self.vocab_size = None
        self._num_feats = None
        self.TGT_DIR = None
        self.INDEX_TO_LABEL = None
        self.LABEL_TO_INDEX = None
        self.train_prefixes = None
        self.valid_prefixes = None
        self.test_prefixes = None
        self.untranscribed_prefixes = None
        self.normalized = False
        self.FEAT_DIR = None
        self.LABEL_DIR = None


    def indices_to_labels(self, indices):
        """ Converts a sequence of indices into their corresponding labels."""

        return [(self.INDEX_TO_LABEL[index]) for index in indices]

    def labels_to_indices(self, labels):
        """ Converts a sequence of labels into their corresponding indices."""

        return [self.LABEL_TO_INDEX[label] for label in labels]

    # TODO Remove all calls to these methods:
    def indices_to_phonemes(self, indices):
        return self.indices_to_labels(indices)
    def phonemes_to_indices(self, phonemes):
        return self.labels_to_indices(phonemes)

    @property
    def num_feats(self):
        """ The number of features per time step in the corpus. """
        if not self._num_feats:
            filename = self.get_train_fns()[0][0]
            feats = np.load(filename)
            # pylint: disable=maybe-no-member
            if len(feats.shape) == 3:
                # Then there are multiple channels of multiple feats
                self._num_feats = feats.shape[1] * feats.shape[2]
            elif len(feats.shape) == 2:
                # Otherwise it is just of shape time x feats
                self._num_feats = feats.shape[1]
            else:
                raise ValueError(
                    "Feature matrix of shape %s unexpected" % str(feats.shape))
        return self._num_feats

    def get_train_fns(self):
        feat_fns = [os.path.join(self.FEAT_DIR, "%s.%s.npy" % (prefix, self.feat_type))
                    for prefix in self.train_prefixes]
        label_fns = [os.path.join(self.LABEL_DIR, "%s.%s" % (prefix, self.label_type))
                      for prefix in self.train_prefixes]
        return feat_fns, label_fns

    def get_valid_fns(self):
        feat_fns = [os.path.join(self.FEAT_DIR, "%s.%s.npy" % (prefix, self.feat_type))
                    for prefix in self.valid_prefixes]
        label_fns = [os.path.join(self.LABEL_DIR, "%s.%s" % (prefix, self.label_type))
                      for prefix in self.valid_prefixes]
        return feat_fns, label_fns

    def get_test_fns(self):
        feat_fns = [os.path.join(self.FEAT_DIR, "%s.%s.npy" % (prefix, self.feat_type))
                    for prefix in self.test_prefixes]
        label_fns = [os.path.join(self.LABEL_DIR, "%s.%s" % (prefix, self.label_type))
                      for prefix in self.test_prefixes]
        return feat_fns, label_fns

    def get_untranscribed_prefixes(self):

        untranscribed_prefix_fn = join(self.TGT_DIR, "untranscribed_prefixes.txt")
        if os.path.exists(untranscribed_prefix_fn):
            with open(untranscribed_prefix_fn) as f:
                prefixes = f.readlines()

            return [prefix.strip() for prefix in prefixes]

        return None

    def get_untranscribed_fns(self):
        feat_fns = [os.path.join(self.FEAT_DIR, "%s.%s.npy" % (prefix, self.feat_type))
                    for prefix in self.untranscribed_prefixes]
        return feat_fns

    def prepare_feats(self, org_dir):
        """ Prepares input features"""

        if not os.path.isdir(self.FEAT_DIR):
            os.makedirs(self.FEAT_DIR)

        should_extract_feats = False
        for fn in os.listdir(org_dir):
            path = join(org_dir, fn)
            if path.endswith(".wav"):
                prefix = os.path.basename(os.path.splitext(path)[0])
                mono16k_wav_path = join(self.FEAT_DIR, "%s.wav" % prefix)
                feat_path = join(self.FEAT_DIR,
                                 "%s.%s.npy" % (prefix, self.feat_type))
                if not os.path.isfile(feat_path):
                    # Then we should extract feats
                    should_extract_feats = True
                    if not os.path.isfile(mono16k_wav_path):
                        feat_extract.convert_wav(path, mono16k_wav_path)

        if should_extract_feats:
            feat_extract.from_dir(self.FEAT_DIR, self.feat_type)

    def get_prefixes(self):
        fns = [fn for fn in os.listdir(self.LABEL_DIR)
               if fn.endswith(self.label_type)]
        prefixes = [os.path.splitext(fn)[0] for fn in fns]
        return prefixes

    def make_data_splits(self, max_samples=1000, seed=0):
        """ Splits the utterances into training, validation and test sets."""

        train_prefix_fn = join(self.TGT_DIR, "train_prefixes.txt")
        valid_prefix_fn = join(self.TGT_DIR, "valid_prefixes.txt")
        test_prefix_fn = join(self.TGT_DIR, "test_prefixes.txt")

        train_f_exists = os.path.isfile(train_prefix_fn)
        valid_f_exists = os.path.isfile(valid_prefix_fn)
        test_f_exists = os.path.isfile(test_prefix_fn)

        if train_f_exists and valid_f_exists and test_f_exists:
            with open(train_prefix_fn) as train_f:
                train_prefixes = [line.strip() for line in train_f]
            with open(valid_prefix_fn) as valid_f:
                valid_prefixes = [line.strip() for line in valid_f]
            with open(test_prefix_fn) as test_f:
                test_prefixes = [line.strip() for line in test_f]

            return train_prefixes, valid_prefixes, test_prefixes

        prefixes = self.get_prefixes()
        # TODO Note that I'm shuffling after sorting; this could be better.
        # TODO Remove explicit reference to "fbank"
        # TODO Surely filtering by size should be done by the corpus reader?
        # How best to keep reads of the corpus consistent?
        prefixes = utils.filter_by_size(
            self.FEAT_DIR, prefixes, "fbank", max_samples)
        Ratios = namedtuple("Ratios", ["train", "valid", "test"])
        # TODO These ratios can't be hardcoded
        ratios = Ratios(.90, .05, .05)
        train_end = int(ratios.train*len(prefixes))
        valid_end = int(train_end + ratios.valid*len(prefixes))
        random.shuffle(prefixes)
        train_prefixes = prefixes[:train_end]
        valid_prefixes = prefixes[train_end:valid_end]
        test_prefixes = prefixes[valid_end:]

        # TODO Perhaps the ReadyCorpus train_prefixes variable should be a
        # property that writes this file when it is changed, then we can remove
        # the code from here. The thing is, the point of those files is that
        # they don't change, so they should be written once only, unless you
        # explicitly delete them. This isn't very clear from the perspective of
        # the user though, so it's a design decision to think about.
        if train_prefixes:
            with open(train_prefix_fn, "w") as train_f:
                for prefix in train_prefixes:
                    print(prefix, file=train_f)
        if valid_prefixes:
            with open(valid_prefix_fn, "w") as dev_f:
                for prefix in valid_prefixes:
                    print(prefix, file=dev_f)
        if test_prefixes:
            with open(test_prefix_fn, "w") as test_f:
                for prefix in test_prefixes:
                    print(prefix, file=test_f)

        return train_prefixes, valid_prefixes, test_prefixes

    def determine_labels(self):
        """ Returns a set of phonemes found in the corpus. """
        phonemes = set()
        for fn in os.listdir(self.LABEL_DIR):
            if fn.endswith(self.label_type):
                with open(join(self.LABEL_DIR, fn)) as f:
                    try:
                        line_phonemes = set(f.readline().split())
                    except UnicodeDecodeError:
                        print("Unicode decode error on file {}".format(fn))
                        raise
                    phonemes = phonemes.union(line_phonemes)
        return phonemes

def check_data(path):
    """ Checks that the data exists and is in the correct directory so that a
    ReadyCorpus can be made out of it."""
    pass

class ReadyCorpus(AbstractCorpus):
    """ Interface to a corpus that has WAV files and label files split into
    utterances and segregated in a directory with a "feat" and "label" dir. """

    def __init__(self, tgt_dir, feat_type="fbank", label_type="phonemes"):
        super().__init__(feat_type, label_type)

        self.TGT_DIR = tgt_dir

        if not os.path.isdir(self.TGT_DIR):
            raise FileNotFoundError("The directory {} does not exist.".format(
                                    self.TGT_DIR))

        self.FEAT_DIR = os.path.join(tgt_dir, "feat")
        self.WAV_DIR = os.path.join(tgt_dir, "wav")
        self.LABEL_DIR = os.path.join(tgt_dir, "label")

        if not os.path.isdir(self.WAV_DIR):
            print(self.WAV_DIR)
            raise PersephoneException("The supplied path requires a 'wav' subdirectory.")
        if not os.path.isdir(self.FEAT_DIR):
            os.makedirs(self.FEAT_DIR)
        if not os.path.isdir(self.LABEL_DIR):
            raise PersephoneException("The supplied path requires a 'label' subdirectory.")

        self.prepare_feats(self.WAV_DIR)
        self.labels = self.determine_labels()
        train, valid, test = self.make_data_splits()

        if train == []:
            print("""WARNING: Corpus object has no training data. Are you sure
            it's in the correct directories? WAVs should be in {} and
            transcriptions in {} with the extension .{}""".format(
                self.WAV_DIR, self.LABEL_DIR, label_type))

        self.train_prefixes = train
        self.valid_prefixes = valid
        self.test_prefixes = test

        # TODO just testing model.transcribe() by using the test set here
        self.untranscribed_prefixes = self.get_untranscribed_prefixes()

        # Sort the training prefixes by size for more efficient training
        self.train_prefixes = utils.sort_by_size(
            self.FEAT_DIR, self.train_prefixes, feat_type)

        # TODO Should be in the abstract corpus. It's common to all corpora but
        # it needs to be set after self.labels. Perhaps I should use a label
        # setter which creates this, then indices_to_phonemes/indices_to_labels
        # will automatically call it.
        self.LABEL_TO_INDEX = {label: index for index, label in enumerate(
                                 ["pad"] + sorted(list(self.labels)))}
        self.INDEX_TO_LABEL = {index: phn for index, phn in enumerate(
                                 ["pad"] + sorted(list(self.labels)))}
        self.vocab_size = len(self.labels)



from setuptools import setup

setup(name='persephone',
      version='0.1.9',
      description='A tool for developing automatic phoneme transcription models',
      long_description=open('README.md').read(),
      url='https://github.com/oadams/persephone',
      author='Oliver Adams',
      author_email='oliver.adams@gmail.com',
      license='GPLv3',
      packages=['persephone', 'persephone.datasets'],
      classifiers = [
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
      ],
      keywords='speech-recognition machine-learning acoustic-models artificial-intelligence neural-networks',
      install_requires=[
           'ipython==6.2.1',
           'GitPython==2.1.8',
           'nltk==3.2.5',
           'numpy==1.14.0',
           'python-speech-features==0.6',
           'scipy==1.0.0',
           'tensorflow==1.4.1',
           'scikit-learn==0.19.1',
           'pympi-ling==1.69',
           'pylint==1.8.2',
      ],
)



""" Miscellaneous utility functions. """
import os
import subprocess
from subprocess import PIPE

from git import Repo
import numpy as np
from nltk.metrics import distance

from . import config
from .exceptions import DirtyRepoException

def is_git_directory_clean(path_to_repo, search_parent_dirs=True):
    """
    Check that the git working directory is in a clean state
    and raise exceptions if not.
    :path_to_repo: The path of the git repo
    """
    repo = Repo(path_to_repo, search_parent_dirs)
    if repo.untracked_files:
        raise DirtyRepoException("Untracked files. Commit them first.")
    # If there are changes to already tracked files
    if repo.is_dirty():
        raise DirtyRepoException("Changes to the index or working tree."
                                 "Commit them first .")

def target_list_to_sparse_tensor(target_list):
    """ Make tensorflow SparseTensor from list of targets, with each element in
    the list being a list or array with the values of the target sequence
    (e.g., the integer values of a character map for an ASR target string) See
    https://github.com/tensorflow/tensorflow/blob/master/tensorflow/
    contrib/ctc/ctc_loss_op_test.py for example of SparseTensor format
    """
    indices = []
    vals = []
    for t_i, target in enumerate(target_list):
        for seq_i, val in enumerate(target):
            indices.append([t_i, seq_i])
            vals.append(val)
    shape = [len(target_list), np.asarray(indices).max(0)[1]+1]
    return (np.array(indices), np.array(vals), np.array(shape))

def zero_pad(matrix, to_length):
    """ Zero pads along the 0th dimension to make sure the utterance array
    x is of length to_length."""

    assert matrix.shape[0] <= to_length
    result = np.zeros((to_length,) + matrix.shape[1:])
    result[:matrix.shape[0]] = matrix
    return result

def collapse(batch_x, time_major=False):
    """ Converts timit into an array of format (batch_size, freq x num_deltas,
    time). Essentially, multiple channels are collapsed to one. """

    new_batch_x = []
    for utterance in batch_x:
        swapped = np.swapaxes(utterance, 0, 1)
        concatenated = np.concatenate(swapped, axis=1)
        new_batch_x.append(concatenated)
    new_batch_x = np.array(new_batch_x)
    if time_major:
        new_batch_x = np.transpose(new_batch_x, (1, 0, 2))
    return new_batch_x

def load_batch_x(path_batch, flatten, time_major=False):
    """ Loads a batch given a list of filenames to numpy arrays in that batch."""

    utterances = [np.load(path) for path in path_batch]
    # The maximum length of an utterance in the batch
    utter_lens = [utterance.shape[0] for utterance in utterances]
    max_len = max(utter_lens)
    batch_size = len(path_batch)
    shape = (batch_size, max_len) + tuple(utterances[0].shape[1:])
    batch = np.zeros(shape)
    for i, utt in enumerate(utterances):
        batch[i] = zero_pad(utt, max_len)
    if flatten:
        batch = collapse(batch, time_major=time_major)
    return batch, np.array(utter_lens)

def batch_per(hyps, refs):
    """ Calculates the phoneme error rate of a batch."""

    total_per = 0
    for i in range(len(hyps)):
        ref = [phn_i for phn_i in refs[i] if phn_i != 0]
        hyp = [phn_i for phn_i in hyps[i] if phn_i != 0]
        total_per += distance.edit_distance(ref, hyp)/len(ref)
    return total_per/len(hyps)

def get_prefixes(dirname, extension):
    """ Returns a list of prefixes to files in the directory (which might be a whole
    corpus, or a train/valid/test subset. The prefixes include the path leading
    up to it, but only the filename up until the first observed period '.'
    """

    prefixes = []
    for root, _, filenames in os.walk(dirname):
        for filename in filenames:
            if filename.endswith(extension):
                # Then it's an input feature file and its prefix will
                # correspond to a training example
                prefixes.append(os.path.join(root, filename.split(".")[0]))
    return sorted(prefixes)

def trim_wav(in_fn, out_fn, start_time, end_time):
    """ Crops the wav file at in_fn so that the audio between start_time and
    end_time is output to out_fn.
    """

    if not os.path.isfile(out_fn):
        args = [config.SOX_PATH, in_fn, out_fn, "trim", str(start_time), "=" + str(end_time)]
        print(args[1:])
        subprocess.run(args)

def make_parent(file_path):
    """ Makes parent dir for a file path."""
    parent_dir = os.path.dirname(file_path)
    if not os.path.isdir(parent_dir):
        os.makedirs(parent_dir)

def get_prefix_lens(feat_dir, prefixes, feat_type):
    prefix_lens = []
    for prefix in prefixes:
        path = os.path.join(feat_dir, "%s.%s.npy" % (prefix, feat_type))
        _, batch_x_lens = load_batch_x([path], flatten=False)
        prefix_lens.append((prefix, batch_x_lens[0]))
    return prefix_lens

def filter_by_size(feat_dir, prefixes, feat_type, max_samples):
    """ Sorts the files by their length and returns those with less
    than or equal to max_samples length. Returns the filename prefixes of
    those files. The main job of the method is to filter, but the sorting
    may give better efficiency when doing dynamic batching unless it gets
    shuffled downstream.
    """

    # TODO Tell the user what utterances we are removing.
    prefix_lens = get_prefix_lens(feat_dir, prefixes, feat_type)
    prefixes = [prefix for prefix, length in prefix_lens
                if length <= max_samples]
    return prefixes

def sort_by_size(feat_dir, prefixes, feat_type):
    prefix_lens = get_prefix_lens(feat_dir, prefixes, feat_type)
    prefix_lens.sort(key=lambda prefix_len: prefix_len[1])
    prefixes = [prefix for prefix, _ in prefix_lens]
    return prefixes

def is_number(string):
    """ Tests if a string is valid float. """
    try:
        float(string)
        return True
    except ValueError:
        return False

def remove_multi(to_remove, target_list):
    """ Removes instances of an item from a list."""
    return list(filter(lambda x: x != to_remove, target_list))

def wav_length(fn):
    """ Returns the length of the WAV file in seconds."""

    args = [config.SOX_PATH, fn, "-n", "stat"]
    p = subprocess.Popen(
        [config.SOX_PATH, fn, "-n", "stat"], stdin=PIPE, stdout=PIPE, stderr=PIPE)
    length_line = str(p.communicate()[1]).split("\\n")[1].split()
    print(length_line)
    assert length_line[0] == "Length"
    return float(length_line[-1])

def calc_time(wav_paths):
    """ Calculates the total spoken time a given number of utterances
    corresponds to. """

    import scipy.io.wavfile as wav

    total_secs = 0
    for path in wav_paths:
        print(path)
        rate, sig = wav.read(path)
        total_secs += (len(sig) / rate)

    total_mins = total_secs / 60
    return total_mins



"""Exceptions that can be raised by the Persephone library are defined here"""

class PersephoneException(Exception):
    """Base class for all exceptions raised by the Persephone library"""

class DirtyRepoException(Exception):
    """
    An exception that is raised if the current working directory
    is in a dirty state according to Git.
    """



"""
Provides a Corpus interface to Kunwinjku data (the Butcher corpus for now;
Steven's data soon).
"""

from collections import namedtuple
import os
from os.path import join
import random


from .. import config
from .. import corpus
from .. import feat_extract
from .. import utils

TGT_DIR = join(config.TGT_DIR, "butcher/kun")
LABEL_DIR = join(TGT_DIR, "labels")
FEAT_DIR = join(TGT_DIR, "feats")
#prepare_butcher_feats("fbank", feat_dir)
#prepare_butcher_labels(label_dir)

def prepare_butcher_labels(label_dir=LABEL_DIR):
    """ Prepares target labels as phonemes """
    # TODO offer label format that is consistent with Steven's
    # data; perhaps by using the orthographic form and lowercasing.

    ## TODO: change import to use
    # from pympi.praat import TextGrid
    from nltk_contrib.textgrid import TextGrid # pylint: disable=import-error

    if not os.path.isdir(label_dir):
        os.makedirs(label_dir)

    for fn in os.listdir(config.BUTCHER_DIR):
        path = join(config.BUTCHER_DIR, fn)
        if path.endswith(".TextGrid"):
            prefix = os.path.basename(os.path.splitext(path)[0])
            out_path = join(label_dir, prefix+".phonemes")
            with open(path) as f, open(out_path, "w") as out_f:
                tg = TextGrid(f.read())
                for tier in tg:
                    if tier.nameid == "etic":
                        transcript = [text for _, _, text in tier.simple_transcript
                                      if text != "<p:>"]
                        print(" ".join(transcript).strip(), file=out_f)

def get_prefixes():
    """ Gets the prefixes for utterances in this corpus."""

    fns = [prefix for prefix in os.listdir(config.BUTCHER_DIR)
           if prefix.endswith(".TextGrid")]
    prefixes = [os.path.splitext(fn)[0] for fn in fns]
    return prefixes

def filter_unlabelled(prefixes, label_type):
    filtered_prefixes = []
    for prefix in prefixes:
        with open(join(LABEL_DIR, "%s.%s" % (prefix, label_type))) as f:
            if len(f.readline().split()) != 0:
                filtered_prefixes.append(prefix)
            else:
                print('The labels of utterance "{prefix}" '\
                      'with label type "{label_type}" are empty. Removed ' \
                      'from utterance set.'.format(prefix=prefix,
                                               label_type=label_type))
    return filtered_prefixes

class Corpus(corpus.AbstractCorpus):
    """ Interface to the Kunwinjku data. """

    # TODO Do I make these attributes non-caps?
    FEAT_DIR = FEAT_DIR
    LABEL_DIR = LABEL_DIR

    def __init__(self, feat_type="fbank", label_type="phonemes"):
        super().__init__(feat_type, label_type)

        # TODO This will break because of undefined variables.
        self.labels = prepare_butcher_phonemes(LABEL_DIR) #pylint: disable=undefined-variable
        train, valid, test = make_data_splits(LABEL_DIR) #pylint: disable=undefined-variable

        # Filter out prefixes that have no transcription. It's probably better
        # to have this after the splitting between train/valid/test sets,
        # because having it before means minor changes in the way labelling
        # occurs would lead to drastically different training sets.
        self.train_prefixes = filter_unlabelled(train, label_type)
        self.valid_prefixes = filter_unlabelled(valid, label_type)
        self.test_prefixes = filter_unlabelled(test, label_type)

        # Sort the training prefixes by size for more efficient training
        self.train_prefixes = utils.sort_by_size(
            self.FEAT_DIR, self.train_prefixes, feat_type)

        # TODO Should be in the abstract corpus. It's common to all corpora but
        # it needs to be set after self.labels. Perhaps I should use a label
        # setter which creates this, then indices_to_phonemes/indices_to_labels
        # will automatically call it.
        self.LABEL_TO_INDEX = {label: index for index, label in enumerate(
                                 ["pad"] + sorted(list(self.labels)))}
        self.INDEX_TO_LABEL = {index: phn for index, phn in enumerate(
                                 ["pad"] + sorted(list(self.labels)))}
        self.vocab_size = len(self.labels)



""" A driver script that runs experiments. """

import logging
import os
from os.path import join
import shutil
import sys

from git import Repo

from . import config
from . import rnn_ctc
from .datasets import na
#from . import datasets.griko
from .datasets import chatino
from .datasets import butcher
#from . import datasets.timit
#from . import datasets.japhug
#from . import datasets.babel
from .corpus_reader import CorpusReader
from .exceptions import PersephoneException, DirtyRepoException
from .utils import is_git_directory_clean

EXP_DIR = config.EXP_DIR

def get_exp_dir_num(parent_dir):
    """ Gets the number of the current experiment directory."""
    return max([int(fn.split(".")[0])
                for fn in os.listdir(parent_dir) if fn.split(".")[0].isdigit()]
                    + [-1])

def _prepare_directory(directory_path):
    """
    Prepare the directory structure required for the experiement
    :returns: returns the name of the newly created directory
    """
    exp_num = get_exp_dir_num(directory_path)
    exp_num = exp_num + 1
    exp_dir = os.path.join(directory_path, str(exp_num))
    if not os.path.isdir(exp_dir):
        os.makedirs(exp_dir)
    return exp_dir

def prep_sub_exp_dir(parent_dir):
    """ Prepares an experiment subdirectory
    :parent_dir: the parent directory
    :returns: returns the name of the newly created subdirectory
    """
    return _prepare_directory(parent_dir)

def prep_exp_dir(directory=EXP_DIR):
    """ Prepares an experiment directory by copying the code in this directory
    to it as is, and setting the logger to write to files in that directory.
    Copies a git hash of the most changes at git HEAD into the directory to
    keep the experiment results in sync with the version control system.
    :directory: The path to directory we are preparing for the experiment,
                which will be created if it does not currently exist.
    :returns: The name of the newly created experiment directory.
    """
    if not os.path.isdir(directory):
        os.makedirs(directory)
    exp_dir = _prepare_directory(directory)
    repo = Repo(".", search_parent_directories=True)
    with open(os.path.join(exp_dir, "git_hash.txt"), "w") as f:
        print("SHA1 hash: {hexsha}".format(hexsha=repo.head.commit.hexsha), file=f)

    return exp_dir


def run():
    """
    The only function that should be called from this module. Ensures
    experiments are documented in their own dir with a reference to the hash
    of the commit used to run the experiment.
    """
    is_git_directory_clean(".")
    exp_dir = prep_exp_dir()
    scaling_graph(exp_dir)
    #rerun_storyfoldxv(exp_dir)

def scaling_graph(exp_dir):

    num_runs = 3
    num_trains = [128,256,512,1024,2048]
    #feat_types = ["fbank_and_pitch", "fbank"]
    #label_types = ["phonemes", "phonemes_and_tones"]

    feat_types = ["fbank", "pitch", "fbank_and_pitch", "phonemes_onehot"]
    label_types = ["tones_notgm"]

    for feat_type in feat_types:
        for label_type in label_types:
            for num_train in num_trains:
                long_exp_dir = os.path.join(exp_dir, feat_type, label_type,
                                            str(num_train))
                os.makedirs(long_exp_dir)
                for i in range(num_runs):
                    train(long_exp_dir, "na", feat_type, label_type, 3, 250,
                          train_rec_type="text", num_train=num_train,
                          max_train_ler=0.7, max_valid_ler=0.7)

def rerun_storyfoldxv(exp_dir):

    valid_test = [('crdo-NRU_F4_RENAMING', 'crdo-NRU_F4_HOUSEBUILDING')]#,
    #              ('crdo-NRU_F4_TRADER_AND_HIS_SON', 'crdo-NRU_F4_RENAMING'),
    #              ('crdo-NRU_F4_ELDERS3', 'crdo-NRU_F4_BURIEDALIVE3'),
    #              ('crdo-NRU_F4_BURIEDALIVE2', 'crdo-NRU_F4_CARAVANS')]

    with open(join(exp_dir, "storyfold_crossvalidation.txt"), "w") as f:
        for valid_text, test_text in valid_test:
            for out_f in [f, sys.stdout]:
                print("Test text: %s" % test_text, file=out_f)
                print("Valid text: %s" % valid_text, file=out_f)
                print("", file=out_f, flush=True)

            train(exp_dir, "na", "fbank_and_pitch", "phonemes_and_tones", 3, 400,
                   valid_story=valid_text, test_story=test_text,
                   max_valid_ler=0.5, max_train_ler=0.1)

def story_fold_cross_validation(exp_dir):

    label_type = "phonemes_and_tones"
    with open(join(exp_dir, "storyfold_crossvalidation.txt"), "w") as f:
        texts = na.get_stories(label_type)
        for i, test_text in enumerate(texts):
            valid_text = texts[(i+1) % len(texts)]
            for out_f in [f, sys.stdout]:
                print(i, file=out_f)
                print("Test text: %s" % test_text, file=out_f)
                print("Valid text: %s" % valid_text, file=out_f)
                print("", file=out_f, flush=True)

            train(exp_dir, "na", "fbank_and_pitch", label_type, 3, 400,
                   valid_story=valid_text, test_story=test_text)

def train(exp_dir, language, feat_type, label_type,
          num_layers, hidden_size,
          num_train=None,
          train_rec_type="text_and_wordlist",
          valid_story=None, test_story=None,
          min_epochs=30, max_valid_ler=1.0, max_train_ler=0.8):
    """ Run an experiment. """

    sub_exp_dir = prep_sub_exp_dir(exp_dir)
    print(sub_exp_dir)

    ## get TF logger
    #log = logging.getLogger('tensorflow')
    #log.setLevel(logging.DEBUG)
    ## create formatter and add it to the handlers
    #formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ## create file handler which logs even debug messages
    #fh = logging.FileHandler(os.path.join(sub_exp_dir, 'tensorflow.log'))
    #fh.setLevel(logging.DEBUG)
    #fh.setFormatter(formatter)
    #log.addHandler(fh)

    if language == "chatino":
        corpus = chatino.Corpus(feat_type, label_type)
    elif language == "na":
        corpus = na.Corpus(feat_type, label_type,
                                    train_rec_type=train_rec_type,
                                    valid_story=valid_story,
                                    test_story=test_story)
    elif language == "kunwinjku":
        # TODO How to choose between the Bird and Butcher corpora?
        corpus = butcher.Corpus(feat_type, label_type)
    else:
        raise PersephoneException("Language '%s' not supported." % language)


    if num_train:
        corpus_reader = CorpusReader(corpus, num_train=num_train)
    else:
        corpus_reader = CorpusReader(corpus)
    print(corpus_reader)
    model = rnn_ctc.Model(sub_exp_dir, corpus_reader,
                          num_layers=num_layers,
                          hidden_size=hidden_size,
                          decoding_merge_repeated=(False if
                                                   label_type=="tones"
                                                   else True))
    model.train(min_epochs=min_epochs,
                max_valid_ler=max_valid_ler,
                max_train_ler=max_train_ler)

    try:
        with open(os.path.join(sub_exp_dir, "train_desc2.txt"), "w") as desc_f:
            for f in [desc_f, sys.stdout]:
                print("language: %s" % language, file=f)
                print("feat_type: %s" % feat_type, file=f)
                print("label_type: %s" % label_type, file=f)
                print("train_rec_type: %s" % train_rec_type, file=f)
                print("num_layers: %d" % num_layers, file=f)
                print("hidden_size: %d" % hidden_size, file=f)
                if num_train:
                    print("num_train: %d" % num_train, file=f)
                print("train duration: %f" % corpus_reader.calc_time(), file=f)
                print("batch_size: %d" % corpus_reader.batch_size, file=f)
                print("Exp dir:", sub_exp_dir, file=f)
    except:
        print("Issues with my printing train_desc2")


def get_simple_model(exp_dir, corpus):
    num_layers = 2
    hidden_size= 250

    def decide_batch_size(num_train):

        if num_train >= 512:
            batch_size = 16
        elif num_train < 128:
            if num_train < 4:
                batch_size = 1
            else:
                batch_size = 4
        else:
            batch_size = num_train / 32

        return batch_size

    batch_size = decide_batch_size(len(corpus.train_prefixes))

    corpus_reader = CorpusReader(corpus, batch_size=batch_size)
    model = rnn_ctc.Model(exp_dir, corpus_reader,
                          num_layers=num_layers,
                          hidden_size=hidden_size,
                          decoding_merge_repeated=True)

    return model

def train_ready(corpus, directory=EXP_DIR):

    print(directory)

    exp_dir = prep_exp_dir(directory=directory)
    model = get_simple_model(exp_dir, corpus)
    model.train(min_epochs=20, early_stopping_steps=3)
    return exp_dir

def transcribe(model_path, corpus):
    """ Applies a trained model to untranscribed data in a Corpus. """

    exp_dir = prep_exp_dir()
    model = get_simple_model(exp_dir, corpus)
    model.transcribe(model_path)



""" Performs feature extraction of WAV files for acoustic modelling."""

import os
import subprocess

import numpy as np
import python_speech_features
import scipy.io.wavfile as wav

from . import config
from .exceptions import PersephoneException

def extract_energy(rate, sig):
    """ Extracts the energy of frames. """

    mfcc = python_speech_features.mfcc(sig, rate, appendEnergy=True)
    energy_row_vec = mfcc[:, 0]
    energy_col_vec = energy_row_vec[:, np.newaxis]
    return energy_col_vec

def fbank(wav_path, flat=True):
    """ Currently grabs log Mel filterbank, deltas and double deltas."""

    (rate, sig) = wav.read(wav_path)
    fbank_feat = python_speech_features.logfbank(sig, rate, nfilt=40)
    energy = extract_energy(rate, sig)
    feat = np.hstack([energy, fbank_feat])
    delta_feat = python_speech_features.delta(feat, 2)
    delta_delta_feat = python_speech_features.delta(delta_feat, 2)
    all_feats = [feat, delta_feat, delta_delta_feat]
    if flat == False:
        all_feats = np.array(all_feats)
        # Make time the first dimension for easy length normalization padding
        # later.
        all_feats = np.swapaxes(all_feats, 0, 1)
        all_feats = np.swapaxes(all_feats, 1, 2)
    else:
        all_feats = np.concatenate(all_feats, axis=1)

    # Log Mel Filterbank, with delta, and double delta
    feat_fn = wav_path[:-3] + "fbank.npy"
    np.save(feat_fn, all_feats)

def mfcc(wav_path):
    """ Grabs MFCC features with energy and derivates. """

    (rate, sig) = wav.read(wav_path)
    feat = python_speech_features.mfcc(sig, rate, appendEnergy=True)
    delta_feat = python_speech_features.delta(feat, 2)
    all_feats = [feat, delta_feat]
    all_feats = np.array(all_feats)
    # Make time the first dimension for easy length normalization padding later.
    all_feats = np.swapaxes(all_feats, 0, 1)
    all_feats = np.swapaxes(all_feats, 1, 2)

    feat_fn = wav_path[:-3] + "mfcc13_d.npy"
    np.save(feat_fn, all_feats)

def combine_fbank_and_pitch(feat_dir, prefix):

    fbank_fn = os.path.join(feat_dir, prefix + ".fbank.npy")
    fbanks = np.load(fbank_fn)
    pitch_fn = os.path.join(feat_dir, prefix + ".pitch.npy")
    pitches = np.load(pitch_fn)

    # Check that the fbanks are flat
    if len(fbanks.shape) == 3:
        def flatten(feats_3d):
            swapped = np.swapaxes(feats_3d, 0, 1)
            concatenated = np.concatenate(swapped, axis=1)
            return concatenated
        fbanks = flatten(fbanks)
    elif len(fbanks.shape) == 2:
        pass
    else:
        raise PersephoneException("Invalid fbank array shape %s" % (str(fbanks.shape)))

    diff = len(fbanks) - len(pitches)

    # When using python_speech_features fbank extraction and kaldi pitch
    # extraction, there are slightly differing numbers of frames. Usually only
    # by 1 or so. Just pad with zeros to match.
    # TODO This could be a sign that it might just be best to use Kaldi for
    # fbank feature extraction as well (I kind of want to see how using those
    # features goes anyway). But I'm currently keeping it this way for
    # experimental consistency.
    if diff > 2:
        raise PersephoneException("Excessive difference in number of frames. %d" % diff)
    elif diff > 0:
        pitches = np.concatenate((np.array([[0,0]]*(len(fbanks) - len(pitches))), pitches))
    fbank_pitch_feats = np.concatenate((fbanks, pitches), axis=1)

    out_fn = os.path.join(feat_dir, prefix + ".fbank_and_pitch.npy")
    np.save(out_fn, fbank_pitch_feats)

# TODO I'm operating at the directory level for pitch feats, but at the file
# level for other things. Change this? Currently wav normalization occurs
# elsewhere too (in say, datasets.chatin.prepare_feats()). Kaldi expects
# specific wav format, so that should be coupled together with pitch extraction
# here.
def from_dir(dirname, feat_type):
    """ Performs feature extraction from the WAV files in a directory. """

    def all_wavs_processed():
        """
        True if all wavs in the directory have corresponding numpy feature
        file; False otherwise.
        """

        for fn in os.listdir(dirname):
            prefix, ext = os.path.splitext(fn)
            if ext == ".wav":
                if not os.path.exists(
                        os.path.join(dirname, "%s.%s.npy" % (prefix, feat_type))):
                    return False
        return True

    if all_wavs_processed():
        # Then nothing needs to be done here
        return
    # Otherwise, go on and process everything...

    # If pitch features are needed as part of this, extract them
    if feat_type == "pitch" or feat_type == "fbank_and_pitch":
        kaldi_pitch(dirname, dirname)

    # Then apply file-wise feature extraction
    for filename in os.listdir(dirname):
        print("Preparing %s features for %s" % (feat_type, filename))
        path = os.path.join(dirname, filename)
        if path.endswith(".wav"):
            if feat_type == "fbank":
                fbank(path)
            elif feat_type == "fbank_and_pitch":
                fbank(path)
                prefix = os.path.splitext(filename)[0]
                combine_fbank_and_pitch(dirname, prefix)
            elif feat_type == "pitch":
                # Already extracted pitch at the start of this function.
                pass
            elif feat_type == "mfcc13_d":
                mfcc(path)
            else:
                raise PersephoneException("Feature type not found: %s" % feat_type)

def convert_wav(org_wav_fn, tgt_wav_fn):
    """ Converts the wav into a 16bit mono 16000Hz wav."""
    args = [config.FFMPEG_PATH,
            "-i", org_wav_fn, "-ac", "1", "-ar", "16000", tgt_wav_fn]
    subprocess.run(args)


def kaldi_pitch(wav_dir, feat_dir):
    """ Extract Kaldi pitch features. Assumes 16k mono wav files."""

    # Make wav.scp and pitch.scp files
    prefixes = []
    for fn in os.listdir(wav_dir):
        prefix, ext = os.path.splitext(fn)
        if ext == ".wav":
            prefixes.append(prefix)

    wav_scp_path = os.path.join(feat_dir, "wavs.scp")
    with open(wav_scp_path, "w") as wav_scp:
        for prefix in prefixes:
            print(prefix, os.path.join(wav_dir, prefix + ".wav"), file=wav_scp)

    pitch_scp_path = os.path.join(feat_dir, "pitch_feats.scp")
    with open(pitch_scp_path, "w") as pitch_scp:
        for prefix in prefixes:
            print(prefix, os.path.join(feat_dir, prefix + ".pitch.txt"), file=pitch_scp)

    # Call Kaldi pitch feat extraction
    args = [os.path.join(config.KALDI_ROOT, "src/featbin/compute-kaldi-pitch-feats"),
            "scp:%s" % (wav_scp_path), "scp,t:%s" % pitch_scp_path]
    subprocess.run(args)

    # Convert the Kaldi pitch *.txt files to numpy arrays.
    for fn in os.listdir(feat_dir):
        if fn.endswith(".pitch.txt"):
            pitch_feats = []
            with open(os.path.join(feat_dir, fn)) as f:
                for line in f:
                    sp = line.split()
                    if len(sp) > 1:
                        pitch_feats.append([float(sp[0]), float(sp[1])])
            prefix, _ = os.path.splitext(fn)
            out_fn = prefix + ".npy"
            a = np.array(pitch_feats)
            np.save(os.path.join(feat_dir, out_fn), a)
